# Copyright Toyota Research Institute 2019

import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import KFold, cross_val_score
from camd import tqdm

import abc


class HypothesisAgent(metaclass=abc.ABCMeta):
    def __init__(self):
        pass

    @abc.abstractmethod
    def get_hypotheses(self, candidate_data):
        """

        Returns:
            subset of candidate data which represent some
            choice e. g. for the next set of experiments

        """


class QBC:
    """
    Helper class for Uncertainty quantification using
    non-supporting regressors with Query-By-Committee
    """
    def __init__(self, n_members, training_fraction, ml_algorithm=None,
                 ml_algorithm_params=None, test_full_model=True):
        """
        n_members (int): Number of committee members (i.e. models to train)
        training_fraction (float): fraction of data to use in training committee members
        ml_algorithm (sklearn.RegressorMixin): sklearn-style regressor for
            regression algorithm
        ml_algorithm_params (dict): parameters to pass to the algorithm
        """
        self.n_members = n_members
        self.training_fraction = training_fraction
        self.ml_algorithm = ml_algorithm if ml_algorithm else LinearRegression
        self.ml_algorithm_params = ml_algorithm_params if ml_algorithm_params else {}
        self.committee_models = []
        self.trained = False
        self.test_full_model = test_full_model
        self.cv_score = np.nan
        self._X = None
        self._y = None

    def fit(self, X, y):
        self._X, self._y = X, y

        split_X = []
        split_y = []

        for i in range(self.n_members):
            a = np.arange(len(X))
            np.random.shuffle(a)
            indices = a[:int(self.training_fraction * len(X))]
            split_X.append(X.iloc[indices])
            split_y.append(y.iloc[indices])

        self.committee_models = []
        for i in tqdm(list(range(self.n_members))):
            scaler = StandardScaler()
            X = scaler.fit_transform(split_X[i])
            y = split_y[i]
            model = self.ml_algorithm(**self.ml_algorithm_params)
            model.fit(X, y)
            self.committee_models.append([scaler, model])  # Note we're saving the scaler to use in predictions

        self.trained = True

        if self.test_full_model:
            # Get a CV score for an overall model with present dataset
            overall_model = self.ml_algorithm(**self.ml_algorithm_params)
            overall_scaler = StandardScaler()
            _X = overall_scaler.fit_transform(self._X, self._y)
            overall_model.fit(_X, self._y)
            cv_score = cross_val_score(overall_model, _X, self._y,
                                       cv=KFold(5, shuffle=True), scoring='neg_mean_absolute_error')
            self.cv_score = np.mean(cv_score) * -1

    def predict(self, X):
        # Apply the committee of models to candidate space
        committee_predictions = []
        for i in tqdm(list(range(self.n_members))):
            scaler = self.committee_models[i][0]
            model = self.committee_models[i][1]
            _X = scaler.transform(X)
            committee_predictions.append(model.predict(_X))
        stds = np.std(np.array(committee_predictions), axis=0)
        means = np.mean(np.array(committee_predictions), axis=0)
        return means, stds


class RandomAgent(HypothesisAgent):
    """
    Baseline agent: Randomly picks from candidate dataset
    """
    def __init__(self, candidate_data=None, seed_data=None, n_query=1):

        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.n_query = n_query
        self.cv_score = np.nan
        super(RandomAgent, self).__init__()

    def get_hypotheses(self, candidate_data, seed_data=None):
        """

        Args:
            candidate_data (DataFrame): candidate data
            seed_data (DataFrame): seed data

        Returns:
            (List) of indices

        """
        return candidate_data.sample(self.n_query).index.tolist()



import unittest
import os
from camd.utils.data import cache_matrio_data, load_dataframe
from camd import CAMD_CACHE


class DataTest(unittest.TestCase):
    def setUp(self):
        # Remove smallest file and cache
        self.smallfile = "oqmd1.2_exp_based_entries_featurized_v2.pickle"
        self.smallfile_path = os.path.join(CAMD_CACHE, self.smallfile)
        if os.path.isfile(self.smallfile_path):
            os.remove(self.smallfile_path)

    def test_cache(self):
        cache_matrio_data(self.smallfile)
        self.assertTrue(os.path.isfile(self.smallfile_path))

        # Ensure no re-download of existing files in cache
        time_before = os.path.getmtime(self.smallfile_path)
        cache_matrio_data(self.smallfile)

        time_after = os.path.getmtime(self.smallfile_path)
        self.assertEqual(time_after, time_before)

    def test_load_dataframe(self):
        df_name = self.smallfile.rsplit('.', 1)[0]
        dataframe = load_dataframe(df_name)
        self.assertEqual(len(dataframe), 36581)


if __name__ == '__main__':
    unittest.main()



# Copyright Toyota Research Institute 2019

import time
import abc
import json
import os
from copy import deepcopy
from multiprocessing import cpu_count
from collections import OrderedDict

import gpflow
import numpy as np
from qmpy.analysis.thermodynamics.phase import Phase, PhaseData
from camd.analysis import PhaseSpaceAL, ELEMENTS
from camd.agent.base import HypothesisAgent, QBC
from camd.log import camd_traced

from sklearn.gaussian_process.kernels import RBF, ConstantKernel
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.model_selection import cross_val_score, KFold, train_test_split
from sklearn.ensemble.bagging import BaggingRegressor
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.ensemble import AdaBoostRegressor
from sklearn.cluster import MiniBatchKMeans
from sklearn.metrics import pairwise_distances

# TODO: Adaptive N_query and subsampling of candidate space
# TODO: Exploit/Explore general method
# TODO: General uncertainty handling?


class StabilityAgent(HypothesisAgent, metaclass=abc.ABCMeta):
    """
    The StabilityAgent is a mixin abstract class which contains
    initialization parameters and methods common to every agent
    which is responsible for making decisions about stability.
    """
    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
                 hull_distance=0.0, multiprocessing=True):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            multiprocessing (bool, int): whether to use multiprocessing
                for phase stability analysis, if an int, sets the n_jobs
                parameter as well.  If a bool, sets n_jobs to cpu_count()
                if True and n_jobs to 1 if false.
        """
        super().__init__()
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.n_query = n_query
        self.hull_distance = hull_distance
        self.pd = None

        # TODO: Probably should just use a single parameter here
        self.multiprocessing = multiprocessing
        if isinstance(self.multiprocessing, bool):
            if self.multiprocessing:
                self.n_jobs = cpu_count()
            else:
                self.n_jobs = 1
        elif isinstance(self.multiprocessing, int) and self.multiprocessing > 0:
            self.n_jobs = self.multiprocessing
        else:
            self.n_jobs = 1

        # These might be able to go into the base class
        self.cv_score = np.nan
        self.indices_to_compute = None

    def get_pd(self):
        """
        Refresh the phase diagram associated with the seed_data

        Returns:
            None
        """
        self.pd = PhaseData()
        phases = [Phase(row['Composition'], energy=row['delta_e'],
                        per_atom=True, description=row_index)
                  for row_index, row in self.seed_data.iterrows()]
        phases.extend([Phase(el, 0.0, per_atom=True) for el in ELEMENTS])
        self.pd.add_phases(phases)
        return self.pd

    def update_data(self, candidate_data=None, seed_data=None):
        """
        Helper function to update the data according to the schema
        of the default OQMD data.  Updates the candidate_data and
        seed_data attributes, and returns the processed features
        and targets associated with the candidates and seed data.

        Args:
            candidate_data (DataFrame): new candidate dataframe
            seed_data (DataFrame): new seed dataframe

        Returns:
            (DataFrame): candidate features
            (DataFrame): seed features
            (DataFrame): seed targets

        """
        # Note: In the drop command, we're ignoring errors for
        #   brevity.  We should watch this, because we may not
        #   drop everything we intend to.
        drop_columns = ['Composition', 'N_species', 'delta_e',
                        'pred_delta_e', 'pred_stability']
        if candidate_data is not None:
            self.candidate_data = candidate_data
            X_cand = candidate_data.drop(
                drop_columns, axis=1, errors='ignore')
        else:
            X_cand = None
        if seed_data is not None:
            self.seed_data = seed_data
            X_seed = self.seed_data.drop(
                drop_columns, axis=1, errors='ignore')
            y_seed = self.seed_data['delta_e']
        else:
            X_seed, y_seed = None, None

        return X_cand, X_seed, y_seed

    def update_candidate_stabilities(self, formation_energies,
                                     sort=True, floor=-6.0):
        """
        Updates the candidate dataframe with the stabilities
        of the candidate compositions according to the requisite
        phase diagram analysis.

        Args:
            formation_energies ([float]): list of predictions for formation
                energies corresponding to candidate_data ordering
            sort (bool): whether or not to sort final list
            floor (float): a float intended to add a floor to the predicted
                formation energies

        Returns:
            (DataFrame): dataframe corresponding to self.candidate_data
        """
        # Preprocess formation energies with floor
        if floor is not None:
            formation_energies = np.array(formation_energies)
            formation_energies[formation_energies < floor] = floor

        # Update formation energy predictions
        self.candidate_data['pred_delta_e'] = formation_energies

        # Construct candidate phases
        candidate_phases = [
            Phase(data['Composition'], energy=data['pred_delta_e'],
                  per_atom=True, description=m_id)
            for m_id, data in self.candidate_data.iterrows()
        ]

        # Refresh and copy seed PD
        pd_ml = deepcopy(self.get_pd())
        pd_ml.add_phases(candidate_phases)
        space_ml = PhaseSpaceAL(bounds=ELEMENTS, data=pd_ml)

        # Compute and return stabilities
        if self.multiprocessing:
            space_ml.compute_stabilities_multi(candidate_phases)
        else:
            space_ml.compute_stabilities_mod(candidate_phases)

        self.candidate_data['pred_stability'] = \
            [phase.stability for phase in candidate_phases]

        if sort:
            self.candidate_data = self.candidate_data.sort_values('pred_stability')

        return self.candidate_data


@camd_traced
class QBCStabilityAgent(StabilityAgent):
    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
                 hull_distance=0.0, multiprocessing=True, alpha=0.5,
                 training_fraction=0.5, ml_algorithm=None,
                 ml_algorithm_params=None, n_members=10):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            multiprocessing (bool): whether to use multiprocessing
                for phase stability analysis
            training_fraction (float): fraction of data to use for
                training committee members
            alpha (float): weighting factor for the stdev in making
                best-case predictions of the stability
            ml_algorithm (sklearn-style regressor): Regression method
            ml_algorithm_params (dict): parameters to pass to the regression
                method
            n_members (int): number of committee members for the qbc
        """

        super(QBCStabilityAgent, self).__init__(
            candidate_data=candidate_data, seed_data=seed_data,
            n_query=n_query, hull_distance=hull_distance,
            multiprocessing=multiprocessing
        )

        self.alpha = alpha
        self.qbc = QBC(
            n_members=n_members, training_fraction=training_fraction,
            ml_algorithm=ml_algorithm,
            ml_algorithm_params=ml_algorithm_params
        )

    def get_hypotheses(self, candidate_data, seed_data=None,
                       retrain_committee=True):
        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)

        # Retrain committee if untrained or if specified
        if not self.qbc.trained or retrain_committee:
            self.qbc.fit(X_seed, y_seed)
        self.cv_score = self.qbc.cv_score

        # QBC makes predictions for Hf and uncertainty on candidate data
        preds, stds = self.qbc.predict(X_cand)
        expected = preds - stds * self.alpha

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(
            expected, sort=True, floor=-6.0)

        # Find the most stable ones up to n_query within hull_distance
        stability_filter = self.candidate_data['pred_stability'] < self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        self.indices_to_compute = within_hull.head(self.n_query).index.tolist()
        return self.indices_to_compute


@camd_traced
class AgentStabilityML5(StabilityAgent):
    """
    An agent that does a certain fraction of full exploration and
    exploitation in each iteration.  It will exploit a fraction of
    N_query options (frac), and explore the rest of its budget.
    """
    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
                 hull_distance=0.0, multiprocessing=True,
                 ml_algorithm=None, ml_algorithm_params=None,
                 exploit_fraction=0.5):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            multiprocessing (bool): whether to use multiprocessing
                for phase stability analysis
            ml_algorithm (sklearn-style regressor): Regression method
            ml_algorithm_params (dict): parameters to pass to the regression
                method
            exploit_fraction (float): fraction of n_query to assign to
                exploitation hypotheses
        """
        super(AgentStabilityML5, self).__init__(
            candidate_data=candidate_data, seed_data=seed_data,
            n_query=n_query, hull_distance=hull_distance,
            multiprocessing=multiprocessing
        )

        self.ml_algorithm = ml_algorithm
        self.ml_algorithm_params = ml_algorithm_params
        self.exploit_fraction = exploit_fraction

    def get_hypotheses(self, candidate_data, seed_data=None):
        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)
        steps = [('scaler', StandardScaler()), ('ML', self.ml_algorithm(**self.ml_algorithm_params))]
        pipeline = Pipeline(steps)

        cv_score = cross_val_score(pipeline, X_seed, self.seed_data['delta_e'],
                                   cv=KFold(5, shuffle=True), scoring='neg_mean_absolute_error')
        self.cv_score = np.mean(cv_score)*-1
        pipeline.fit(X_seed, self.seed_data['delta_e'])

        expected = pipeline.predict(X_cand)

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(
            expected, sort=True, floor=-6.0)

        # Filter by stability according to hull distance
        stability_filter = self.candidate_data['pred_stability'] < self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        # Exploitation part:
        n_exploitation = int(self.n_query * self.exploit_fraction)
        to_compute = within_hull.head(n_exploitation).index.tolist()
        remaining = within_hull.tail(len(within_hull)- n_exploitation)
        remaining = remaining.append(self.candidate_data[~stability_filter])

        # Exploration part (pick randomly from remainder):
        n_exploration = self.n_query - n_exploitation

        if n_exploration > len(remaining):
            n_exploration = len(remaining)
        to_compute.extend(remaining.sample(n_exploration).index.tolist())

        self.indices_to_compute = to_compute
        return self.indices_to_compute


class GaussianProcessStabilityAgent(StabilityAgent):
    """
    Simple Gaussian Process Regressor based Stability Agent
    """
    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
                 hull_distance=0.0, multiprocessing=True,
                 alpha=0.5):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            multiprocessing (bool): whether to use multiprocessing
                for phase stability analysis
            alpha (float): weighting factor for the stdev in making
                best-case predictions of the stability
        """
        super(GaussianProcessStabilityAgent, self).__init__(
            candidate_data=candidate_data, seed_data=seed_data,
            n_query=n_query, hull_distance=hull_distance,
            multiprocessing=multiprocessing
        )
        self.multiprocessing = multiprocessing
        self.alpha = alpha
        self.GP = GaussianProcessRegressor(kernel=ConstantKernel(1) * RBF(1), alpha=0.002)

    def get_hypotheses(self, candidate_data, seed_data=None):
        X_cand, X_seed, y_seed = self.update_data(candidate_data=candidate_data, seed_data=seed_data)

        steps = [('scaler', StandardScaler()), ('GP', self.GP)]
        cv_pipeline = Pipeline(steps)
        self.cv_score = np.mean(
            -1.0 * cross_val_score(cv_pipeline, X_seed, y_seed, cv=KFold(3, shuffle=True),
                                   scoring='neg_mean_absolute_error'))

        steps = [('scaler', StandardScaler()), ('GP', self.GP)]
        overall_pipeline = Pipeline(steps)
        overall_pipeline.fit(X_seed, y_seed)

        # GP makes predictions for Hf and uncertainty*alpha on candidate data
        preds, stds = overall_pipeline.predict(X_cand, return_std=True)
        expected = preds - stds * self.alpha

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(
            expected, sort=True, floor=-6.0)

        # Find the most stable ones up to n_query within hull_distance
        stability_filter = self.candidate_data['pred_stability'] < self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        self.indices_to_compute = within_hull.head(self.n_query).index.tolist()
        return self.indices_to_compute


class SVGProcessStabilityAgent(StabilityAgent):
    """
    Stochastic variational gaussian process stability agent for Big Data.

    The computational complexity of this algorithm scales as O(M^3)
    compared to O(N^3) of standard GP, where N is the number of data points
    and M is the number of inducing points (M<<N).

    The default parameters are optimized to deliver a compromise between
    compute-time and model accuracy for data sets with up to 25 to 40
    thousand examples (e.g. the ICSD seed data). For bigger systems,
    parameter M may need to be reduced. For smaller systems, it can be
    increased, if higher accuracy is desired.  Inducing point locations
    are determined using k-means clustering.

    References:
        Hensman, James, Nicolo Fusi, and Neil D. Lawrence. "Gaussian
            processes for big data." Uncertainty in Artificial
            Intelligence (2013).
        Kingma, Diederik P., and Jimmy Ba. "Adam: A method for stochastic
            optimization." arXiv preprint arXiv:1412.6980 (2014).

    """
    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
                 hull_distance=0.0, multiprocessing=True,
                 alpha=0.5, M=600):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            multiprocessing (bool): whether to use multiprocessing
                for phase stability analysis
            alpha (float): weighting factor for the stdev in making
                best-case predictions of the stability
            M (int): number of inducing points associated with the
                SVGProcess
        """
        super(SVGProcessStabilityAgent, self).__init__(
            candidate_data=candidate_data, seed_data=seed_data,
            n_query=n_query, hull_distance=hull_distance,
            multiprocessing=multiprocessing
        )
        self.alpha = alpha
        self.M = M

        # Define non-argument SVG-specific attributes
        self.kernel = gpflow.kernels.RBF(273) * gpflow.kernels.Constant(273)
        self.mean_f = gpflow.mean_functions.Constant()
        self.logger = None
        self.model = None
        self.pred_y = None
        self.pred_std = None

    def get_hypotheses(self, candidate_data, seed_data=None):
        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)

        # Test model performance first.  Note we avoid doing CV to
        # reduce compute time. We simply do a 1-way split 80:20 (train:test)
        X_train, X_test, y_train, y_test = train_test_split(X_seed, y_seed, test_size=0.2, random_state=42)
        scaler = StandardScaler()
        X_train_scaled = scaler.fit_transform(X_train)

        # Do a k-means clustering and use cluster centers as inducing points.
        cls = MiniBatchKMeans(n_clusters=self.M, batch_size=200)
        cls.fit(X_train_scaled)
        Z = cls.cluster_centers_
        _y = np.array(y_train.to_list())
        mu = np.mean(_y)
        sig = np.std(_y)
        print(Z, _y.shape)
        print(sig, mu)
        model = gpflow.models.SVGP(
            X_train_scaled, ((_y - mu)/sig).reshape(-1, 1), self.kernel,
            gpflow.likelihoods.Gaussian(), Z, mean_function=self.mean_f,
            minibatch_size=100
        )
        print("training")
        t0 = time.time()
        logger = self.run_adam(model, gpflow.test_util.notebook_niter(20000))
        print("elapsed time: ", time.time()-t0)

        pred_y, pred_v = model.predict_y(scaler.transform(X_test))
        pred_y = pred_y * sig + mu
        self.cv_score = np.mean(np.abs(pred_y - y_test.to_numpy().reshape(-1, 1)))
        print("cv score", self.cv_score)
        self.model = model

        # Overall model
        scaler = StandardScaler()
        X_scaled = scaler.fit_transform(X_seed)
        cls = MiniBatchKMeans(n_clusters=self.M, batch_size=200)
        cls.fit(X_train_scaled)
        Z = cls.cluster_centers_
        _y = np.array(y_seed.to_list())
        mu = np.mean(_y)
        sig = np.std(_y)
        model = gpflow.models.SVGP(
            X_scaled, ((_y - mu)/sig).reshape(-1, 1), self.kernel,
            gpflow.likelihoods.Gaussian(), Z, mean_function=self.mean_f,
            minibatch_size=100
        )
        logger = self.run_adam(model, gpflow.test_util.notebook_niter(20000))
        print(self.model)
        self.model = model

        # GP makes predictions for Hf and uncertainty*alpha on candidate data
        pred_y, pred_v = model.predict_y(scaler.transform(X_cand))
        pred_y = pred_y * sig + mu
        self.pred_y = pred_y.reshape(-1,)
        self.pred_std = (pred_v**0.5).reshape(-1,)

        expected = self.pred_y - self.pred_std * self.alpha
        print("expected improv", expected)

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(
            expected, sort=True, floor=-6.0)

        # Find the most stable ones up to n_query within hull_distance
        stability_filter = self.candidate_data['pred_stability'] < self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        self.indices_to_compute = within_hull.head(self.n_query).index.tolist()
        return self.indices_to_compute

    def run_adam(self, model, iterations):
        """
        Adam optimizer as implemented in:
        https://github.com/GPflow/GPflow/blob/develop/doc/source/notebooks/advanced/gps_for_big_data.ipynb
        """
        # Create an Adam Optimiser action
        adam = gpflow.train.AdamOptimizer().make_optimize_action(model)

        # Create a Logger action
        self.logger = self.Logger(model)
        actions = [adam, self.logger]

        # Create optimisation loop that interleaves Adam with Logger
        loop = gpflow.actions.Loop(actions, stop=iterations)()

        # Bind current TF session to model
        model.anchor(model.enquire_session())
        return self.logger

    class Logger(gpflow.actions.Action):
        """
        Logger class as implemented in
        https://github.com/GPflow/GPflow/blob/develop/doc/source/notebooks/advanced/gps_for_big_data.ipynb
        """
        def __init__(self, model):
            self.model = model
            self.logf = []

        def run(self, ctx):
            if (ctx.iteration % 10) == 0:
                # Extract likelihood tensor from Tensorflow session
                likelihood = - ctx.session.run(self.model.likelihood_tensor)

                # Append likelihood value to list
                self.logf.append(likelihood)


class BaggedGaussianProcessStabilityAgent(StabilityAgent):
    """
    An ensemble GP learner that can handle relatively large
    datasets by bagging. WIP.

    Current strategy is that weak GP learners are trained on
    random subsets of data with max 500 points (and no bootstrapping),
    to learn a stronger, ensemble learner that minimizes model variance.

    Learned model is usually stronger, and on par with Random Forest
    or NN. We assume the uncertainty of the prediction to be
    the minimum std among all GP estimators trained, for a given point.

    """
    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
                 hull_distance=0.0, multiprocessing=True,
                 alpha=0.5, n_estimators=8, max_samples=5000,
                 bootstrap=False):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            multiprocessing (bool): whether to use multiprocessing
                for phase stability analysis
            alpha (float): scaling for stdev for estimation of best-
                case formation energies
        """
        super(BaggedGaussianProcessStabilityAgent, self).__init__(
            candidate_data=candidate_data, seed_data=seed_data,
            n_query=n_query, hull_distance=hull_distance,
            multiprocessing=multiprocessing
        )

        self.alpha = alpha
        self.n_estimators = n_estimators if n_estimators else 8
        self.max_samples = max_samples if max_samples else 5000
        self.bootstrap = bootstrap if bootstrap else False

        self.GP = GaussianProcessRegressor(kernel=ConstantKernel(1) * RBF(1), alpha=0.002)

    def get_hypotheses(self, candidate_data, seed_data=None):
        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)

        steps = [('scaler', StandardScaler()), ('GP', self.GP)]
        pipeline = Pipeline(steps)

        bag_reg = BaggingRegressor(
            base_estimator=pipeline, n_estimators=self.n_estimators,
            max_samples=self.max_samples, bootstrap=self.bootstrap,
            verbose=True, n_jobs=self.n_jobs)
        self.cv_score = np.mean(
            -1.0 * cross_val_score(
                pipeline, X_seed, y_seed, cv=KFold(3, shuffle=True),
                scoring='neg_mean_absolute_error')
        )
        bag_reg.fit(X_seed, y_seed)

        # TODO: make this a static method
        def _get_unc(bagging_regressor, X_test):
            stds = []
            pres = []
            for est in bagging_regressor.estimators_:
                _p, _s = est.predict(X_test, return_std=True)
                stds.append(_s)
                pres.append(_p)
            return np.mean(np.array(pres), axis=0), np.min(np.array(stds), axis=0)

        # GP makes predictions for Hf and uncertainty*alpha on candidate data
        preds, stds = _get_unc(bag_reg, X_cand)
        expected = preds - stds * self.alpha

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(
            expected, sort=True, floor=-6.0)

        # Find the most stable ones up to n_query within hull_distance
        stability_filter = self.candidate_data['pred_stability'] < self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        self.indices_to_compute = within_hull.head(self.n_query).index.tolist()
        return self.indices_to_compute


@camd_traced
class AgentStabilityAdaBoost(StabilityAgent):
    """
    An agent that does a certain fraction of full exploration and
    exploitation in each iteration.  It will exploit a fraction
    of N_query options (frac), and explore the rest of its budget.
    """
    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
                 hull_distance=0.0, multiprocessing=True,
                 ml_algorithm=None, ml_algorithm_params=None,
                 uncertainty=True, alpha=0.5, n_estimators=10,
                 exploit_fraction=0.5, diversify=False, dynamic_alpha=False):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            multiprocessing (bool): whether to use multiprocessing
                for phase stability analysis
            ml_algorithm (sklearn-style regressor): Regression method
            ml_algorithm_params (dict): parameters to pass to the regression
                method
            uncertainty (bool): whether uncertainty is included in
                minimal predictions
            alpha (float): weighting factor for the stdev in making
                best-case predictions of the stability
            n_estimators (int): number of estimators fro the AdaBoosting
                algorithm
            exploit_fraction (float): fraction of n_query to assign to
                exploitation hypotheses
            diversify (bool): Turns on the diversification algorithm for
                selected experiments.
            dynamic_alpha (bool): Turns on a simple linear schedule where the
                alpha (how much uncertainty is mixed into predictions) starts
                from zero, and increases at a rate of 0.1/iter until it's capped
                at parameter alpha specified for the rest of the iterations.
        """

        super(AgentStabilityAdaBoost, self).__init__(
            candidate_data=candidate_data, seed_data=seed_data,
            n_query=n_query, hull_distance=hull_distance,
            multiprocessing=multiprocessing
        )

        self.ml_algorithm = ml_algorithm
        self.ml_algorithm_params = ml_algorithm_params
        self.exploit_fraction = exploit_fraction
        self.uncertainty = uncertainty
        self.alpha = alpha
        self.n_estimators = n_estimators
        self.diversify = diversify
        self.dynamic_alpha = dynamic_alpha

    def get_hypotheses(self, candidate_data, seed_data=None):
        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)

        steps = [('scaler', StandardScaler()), ('ML', self.ml_algorithm(**self.ml_algorithm_params))]
        pipeline = Pipeline(steps)

        adaboost = AdaBoostRegressor(
            base_estimator=pipeline, n_estimators=self.n_estimators)

        cv_score = cross_val_score(
            adaboost, X_seed, y_seed,
            cv=KFold(3, shuffle=True), scoring='neg_mean_absolute_error')
        self.cv_score = np.mean(cv_score)*-1

        # We will take standard scaler out of the pipleine for
        # prediction purposes (we want a single scaler)

        scaler = StandardScaler()
        X_scaled = scaler.fit_transform(X_seed)
        overall_adaboost = AdaBoostRegressor(
            base_estimator=self.ml_algorithm(**self.ml_algorithm_params),
            n_estimators=self.n_estimators
        )
        overall_adaboost.fit(X_scaled, self.seed_data['delta_e'])

        X_cand = scaler.transform(X_cand)
        expected = overall_adaboost.predict(X_cand)

        if self.uncertainty:
            if self.dynamic_alpha and os.path.exists('iteration.json'):
                with open('iteration.json', 'r') as f:
                    iter = json.load(f)
                print("dynamic alpha activated")
                expected -= min(0.1*iter,self.alpha) * self._get_unc_ada(overall_adaboost, X_cand)
            else:
                expected -= self.alpha * self._get_unc_ada(overall_adaboost, X_cand)

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(
            expected, sort=True, floor=-6.0)

        # Filter by stability according to hull distance
        stability_filter = self.candidate_data['pred_stability'] < self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        # Exploitation part:
        n_exploitation = int(self.n_query * self.exploit_fraction)
        if self.diversify:
            to_compute = diverse_quant( within_hull.index.tolist(), n_exploitation, self.candidate_data)
        else:
            to_compute = within_hull.head(n_exploitation).index.tolist()
        remaining = within_hull.tail(len(within_hull)- n_exploitation)
        remaining = remaining.append(self.candidate_data[~stability_filter])

        # Exploration part (pick randomly from remainder):
        n_exploration = self.n_query - n_exploitation

        if n_exploration > len(remaining):
            n_exploration = len(remaining)
        to_compute.extend(remaining.sample(n_exploration).index.tolist())

        self.indices_to_compute = to_compute
        return self.indices_to_compute

    @staticmethod
    def _get_unc_ada(ada, X):
        preds = []
        for i in ada.estimators_:
            preds.append(i.predict(X))
        preds = np.array(preds)
        preds = preds.T
        stds = []
        for i in preds:
            average = np.average(i, weights=ada.estimator_weights_)
            _std = np.sqrt(np.average((i - average) ** 2, weights=ada.estimator_weights_))
            stds.append(_std)
        return np.array(stds)


def diverse_quant(points, target_length, df, quantiles=None):
    """
    Diversify a sublist by eliminating entries based on comparisons
    with quantiles threshold and Euclidean distance.

    This method takes the points list (which would be the object
    within_hull in stability implementations), and selects a diverse
    subset for the number of exploitation choices (target_length).

    It follows a simple algorithm: start from i = 0 of the points,
    and go down the remainder of the list, removing entries that seem
    to be closer to i below a certain distance threshold. It repeats
    this process for all i: i = 1, i = 2 ... i_max.

    The method tries to adjust the distance threshold until it finds
    the shortest resulting list that is longer than target_length. If
    it can't, it will return points as it is. The threshold values are
    decided by finding distances corresponding to quantiles of the a
    sampled distribution of distances in the overall feature set.

    The method does not alter the original ordering in the list points.

    The intuition behind the algorithm is to make risk-averse choices
    by avoiding the acquisition of too similar candidates, in case one
    example among those entries is sufficient for the model to minimize
    its uncertainty and/or make a decision to not acquire any other in
    that region.

    So the resources would not be wasted and can be allocated to other
    promising choices in points.

    Args:
        points (list): Initial set of points reflecting a preferred order
        target_length (int): length of desired sublist
        df (DataFrame): feature vectors of points, where index labels
            contain elements of points
        quantiles (list): quantilies to test for threshold.
            Defaults to [0.01, 0.02, 0.03, 0.04, 0.05]

    Returns:
        A diversified sublist of points

    """
    quantiles = quantiles if quantiles else [0.01, 0.02, 0.03, 0.04, 0.05]
    if target_length >= len(points):
        return points
    if len(df) > 6000:
        _df = df.sample(6000)
    else:
        _df = df
    drop_columns = ['Composition', 'N_species', 'delta_e',
                    'pred_delta_e', 'pred_stability']
    scaler = StandardScaler()
    X = scaler.fit_transform(_df.drop(drop_columns, axis=1, errors='ignore'))
    flatD = pairwise_distances(X).flatten()

    _df2 = df.loc[points]
    X = scaler.transform(_df2.drop(drop_columns, axis=1, errors='ignore'))
    D = pairwise_distances(X)

    remove_len = len(points) - target_length
    res = []
    for alpha in quantiles:
        q = np.quantile(flatD, alpha)
        to_remove = []
        for i in range(0, len(points)):
            for j in range(i+1, len(points)):
                if D[i, j] < q:
                    to_remove.append(j)
        _rl = len(set(to_remove))
        print(_rl, remove_len)
        if _rl >= remove_len:
            break
        res.append(to_remove)
    if len(res) == 0:
        return points[:target_length]
    else:
        d = OrderedDict()
        for i in res[-1]: # fall back to the latest remove list before break
            d[i] = None
        final_remove_list = list(d.keys())

        return [p for p in points if np.where(_df2.index==p)[0][0]
                not in final_remove_list][:target_length]



# Computational Autonomy for Materials Discovery
# Copyright Toyota Research Institute 2019
import os
from functools import partial
from tqdm import tqdm as _tqdm


__version__ = "2020.3.24"

CAMD_ROOT = os.path.dirname(os.path.abspath(__file__))
CAMD_TEST_FILES = os.path.join(CAMD_ROOT, "tests", "test_files")
CAMD_CACHE = os.path.join(CAMD_ROOT, "_cache")

# Environment-based settings
TQDM_OFF = os.environ.get("TQDM_OFF", None)
CAMD_S3_BUCKET = os.environ.get("CAMD_S3_BUCKET", None)
CAMD_STOP_FILE = os.environ.get("CAMD_STOP_FILE", os.path.join(CAMD_ROOT, "stop"))

if TQDM_OFF:
    tqdm = partial(_tqdm, disable=TQDM_OFF)
else:
    tqdm = _tqdm


#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.

import unittest
import subprocess
import boto3
import json
import time
import shlex
from multiprocessing import Pool
from camd import CAMD_S3_BUCKET
from camd.campaigns.worker import Worker


def teardown_s3():
    """Tear down test files in s3"""
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(CAMD_S3_BUCKET)
    bucket.objects.filter(Prefix="{}".format("oqmd-atf")).delete()


class WorkerTest(unittest.TestCase):
    def tearDown(self):
        teardown_s3()
        Worker("oqmd-atf").remove_stop_file()

    def submit_chemsyses(self, chemsyses):
        # Upload three things to s3
        s3_resource = boto3.resource("s3")
        for chemsys in chemsyses:
            key = "oqmd-atf/submit/{}/status.json".format(chemsys)
            obj = s3_resource.Object(CAMD_S3_BUCKET, key)
            obj.put(Body=json.dumps({"last_submitted": 10}))
            time.sleep(2)

    def put_runs(self, chemsyses):
        # Upload three things to s3
        s3_resource = boto3.resource("s3")
        for chemsys in chemsyses:
            key = "oqmd-atf/runs/{}/job_status.json".format(chemsys)
            obj = s3_resource.Object(CAMD_S3_BUCKET, key)
            obj.put(Body=json.dumps({"status": "started"}))

            key = "oqmd-atf/runs/{}/job_status.json".format(chemsys)
            obj = s3_resource.Object(CAMD_S3_BUCKET, key)
            obj.put(Body=json.dumps({"status": "started"}))

    def test_get_latest_chemsys(self):
        self.submit_chemsyses(["O-V", "O-Ti", "Fe-O"])
        worker = Worker("oqmd-atf")
        latest_chemsys = worker.get_latest_chemsys()
        self.assertEqual(latest_chemsys, "Fe-O")

        self.put_runs(["Fe-O"])
        latest_chemsys = worker.get_latest_chemsys()
        self.assertEqual(latest_chemsys, "O-Ti")

        self.put_runs(["O-V", "O-Ti"])
        latest_chemsys = worker.get_latest_chemsys()
        self.assertIsNone(latest_chemsys)

    def test_run_atf_campaign(self):
        self.submit_chemsyses(["O-Ti", "Fe-O"])
        worker = Worker("oqmd-atf")

        latest_chemsys = worker.get_latest_chemsys()
        self.assertEqual(latest_chemsys, "Fe-O")

        worker.start(num_loops=1)
        latest_chemsys = worker.get_latest_chemsys()
        self.assertEqual(latest_chemsys, "O-Ti")

        worker.start(num_loops=1)
        latest_chemsys = worker.get_latest_chemsys()
        self.assertIsNone(latest_chemsys)

    # TODO: This test is super slow, could make a
    #  new dummy campaign that executes with smaller overhead
    def test_stop(self):
        # Stopping a-priori
        worker = Worker("oqmd-atf")
        worker.write_stop_file()
        executed = worker.start()
        self.assertEqual(executed, 0)

        # # Ensure restarts after stop removal
        self.submit_chemsyses(["O-Ti", "Fe-O"])
        worker = Worker("oqmd-atf")
        worker.remove_stop_file()
        executed = worker.start(num_loops=2)
        self.assertEqual(executed, 2)

        # Ensure restarts after stop removal
        worker = Worker("oqmd-atf")
        worker.remove_stop_file()

        # TODO: this is a pretty hackish way of executing
        #  these in parallel and isn't guaranteed to work,
        #  but works for now

        with Pool(2) as p:
            result = p.map(worker_process, [0, 1])
        self.assertEquals(result[0], 1)

    def test_cli(self):
        self.submit_chemsyses(["O-Ti"])
        output = subprocess.check_output(
            "camd_worker start --campaign oqmd-atf --loops 1",
            shell=True
        )
        self.assertIn("Running experiments", output.decode('utf-8'))


def worker_process(index):
    if index == 0:
        print("index 0")
        worker = Worker("oqmd-atf")
        latest = worker.get_latest_chemsys()
        result = worker.start(sleep_time=7)
        print("returning {} {}".format(result, latest))
        return result
    else:
        time.sleep(3)
        worker = Worker("oqmd-atf")
        print("writing stop file")
        worker.write_stop_file()
        return None


if __name__ == '__main__':
    unittest.main()



"""
Preliminary module for determining search spaces
"""
import os
import pandas as pd
import abc
import warnings
import itertools
import numpy as np
import uuid

from camd import CAMD_CACHE
from camd.utils.data import cache_matrio_data

try:
    from protosearch.build_bulk.oqmd_interface import OqmdInterface
except ModuleNotFoundError:
    print("No module found for camd.domain")


from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen import Composition, Element
from matminer.featurizers.base import MultipleFeaturizer
from matminer.featurizers.composition import ElementProperty, Stoichiometry, ValenceOrbital, IonProperty
from matminer.featurizers.structure import (SiteStatsFingerprint, StructuralHeterogeneity,
                                            ChemicalOrdering, StructureComposition, MaximumPackingEfficiency)


class DomainBase(abc.ABC):
    """
    Domains combine generation and featurization and prepare
    the search space for CAMD Loop.
    """
    @abc.abstractmethod
    def candidates(self):
        """
        Primary method for every Domain to provide candidates.

        Returns:
            (pandas.DataFrame): features for generated hypothetical
                structures. The Index of dataframe should be the
                unique ids for the structures.
        """
        pass

    @property
    @abc.abstractmethod
    def bounds(self):
        """
        Returns:
            list: names of dimensions of the search space.
        """
        pass

    @abc.abstractmethod
    def sample(self, num_samples):
        """
        Abstract method for sampling from created domain
        Args:
            num_samples:
        Returns:
        """
        pass

    @property
    def bounds_string(self):
        """
        Returns: a string representation of search space bounds: e.g. "Ir-Fe-O" or "x1-x2-x3"
        """
        return '-'.join(self.bounds)


class StructureDomain(DomainBase):
    """
    Provides machine learning ready candidate domains (search spaces) for
    hypothetical structures.  If scanning an entire system, use the
    StructureDomain.from_bounds method. If scanning for formula(s),
    provide a list of formulas directly to StructureDomain.

    Once the StructureDomain is initialized, the method candidates returns
    a fully-featurized hypothetical materials set subject to n_max_atoms.

    Args:
        formulas ([str]): list of chemical formulas to create new
            material candidates.
    """
    def __init__(self, formulas, n_max_atoms=None):
        self.formulas = formulas
        self.n_max_atoms = n_max_atoms
        self.features = None
        self._hypo_structures = None

    @classmethod
    def from_bounds(cls, bounds, n_max_atoms=None, charge_balanced=True,
                    create_subsystems=False, **kwargs):
        """
        Convenience constructor that delivers an ML-ready domain from defined chemical boundaries.
        Args:
            bounds:
            charge_balanced:
            frequency_threshold:
            create_subsystems:
            kwargs: arguments to pass to formula creator
        """
        formulas = create_formulas(bounds, charge_balanced=charge_balanced,
                                   create_subsystems=create_subsystems, **kwargs)
        print("Generated chemical formulas: {}".format(formulas))
        return cls(formulas, n_max_atoms)

    @property
    def bounds(self):
        bounds = set()
        for formula in self.formulas:
            bounds = bounds.union(Composition(formula).as_dict().keys())
        return bounds

    def get_structures(self):
        """
        Method to call the external structure generator.
        """
        if self.formulas:
            print("Generating hypothetical structures...")
            self._hypo_structures = get_structures_from_protosearch(self.formulas)
            print("Generated {} hypothetical structures".format(len(self.hypo_structures)))
        else:
            raise ValueError("Need formulas to create structures")

    @property
    def hypo_structures(self):
        """
        Returns (dataframe): Hypothetical structures generated by
            protosearch, filtered by n_max_atoms

        """
        if self._hypo_structures is None:
            self.get_structures()
        if self.n_max_atoms:
            n_max_filter = [i.num_sites <= self.n_max_atoms for i in self._hypo_structures['pmg_structures']]
            if self._hypo_structures is not None:
                return self._hypo_structures[n_max_filter]
            else:
                return None
        else:
            return self._hypo_structures

    @property
    def hypo_structures_dict(self):
        """
        Returns:
            (dict): Hypothetical structures generated by
                protosearch, filtered by n_max_atoms

        """
        return self.hypo_structures["pmg_structures"].to_dict()

    @property
    def compositions(self):
        """
        Returns:
            (list): Compositions of hypothetical structures generated.
        """
        if self.hypo_structures is not None:
            return [s.composition for s in self.hypo_structures]
        else:
            warnings.warn("No stuctures available.")
            return []

    @property
    def formulas_with_valid_structures(self):
        # Note the redundancy here is for pandas to work
        if self.valid_structures is not None:
            return [s.composition.formula
                    for s in self.valid_structures["pmg_structures"]]
        else:
            warnings.warn("No structures available yet.")
            return []

    def featurize_structures(self, featurizer=None, **kwargs):
        """
        Featurizes the hypothetical structures available from
        hypo_structures method. Hypothetical structures for which
        featurization fails is removed and valid structures are
        made available as valid_structures

        Args:
            featurizer (Featurizer): A MatMiner Featurizer.
                Defaults to MultipleFeaturizer with PRB Ward
                Voronoi descriptors.
            **kwargs (dict): kwargs passed to featurize_many
                method of featurizer.

        Returns:
            pandas.DataFrame: features
        """
        # Note the redundancy here is for pandas to work
        if self.hypo_structures is None:
            warnings.warn("No structures available. Generating structures.")
            self.get_structures()

        print("Generating features")
        featurizer = featurizer if featurizer else MultipleFeaturizer([
            SiteStatsFingerprint.from_preset("CoordinationNumber_ward-prb-2017"),
            StructuralHeterogeneity(),
            ChemicalOrdering(),
            MaximumPackingEfficiency(),
            SiteStatsFingerprint.from_preset("LocalPropertyDifference_ward-prb-2017"),
            StructureComposition(Stoichiometry()),
            StructureComposition(ElementProperty.from_preset("magpie")),
            StructureComposition(ValenceOrbital(props=['frac'])),
            StructureComposition(IonProperty(fast=True))
        ])

        features = featurizer.featurize_many(self.hypo_structures['pmg_structures'],
                                             ignore_errors=True, **kwargs)

        n_species, formula = [], []
        for s in self.hypo_structures['pmg_structures']:
            n_species.append(len(s.composition.elements))
            formula.append(s.composition.formula)

        self._features_df = pd.DataFrame.from_records(features, columns=featurizer.feature_labels())
        self._features_df.index = self.hypo_structures.index
        self._features_df['N_species'] = n_species
        self._features_df['Composition'] = formula
        self.features = self._features_df.dropna(axis=0, how='any')
        self.features = self.features.reindex(sorted(self.features.columns), axis=1)

        self._valid_structure_labels = list(self.features.index)
        self.valid_structures = self.hypo_structures.loc[self._valid_structure_labels]

        print("{} out of {} structures were successfully featurized.".format(self.features.shape[0],
                                                                             self._features_df.shape[0]))
        return self.features

    def candidates(self, include_composition=True):
        """
        This is the recommended convenience method that returns
        a fully-featurized set of hypothetical structures.

        Args:
            include_composition (bool): Adds a column named "formula"
                to the dataframe.

        Returns:
            (pandas.DataFrame) feature vectors of valid
                hypothetical structures.
        """
        if self._hypo_structures is None:
            self.get_structures()

        if self.features is None:
            self.featurize_structures()
        if include_composition:
            return self.features
        else:
            return self.features.drop('Composition', axis=1)

    def sample(self, num_samples):
        self.candidates().sample(num_samples)


def get_structures_from_protosearch(formulas, source='icsd',
                                    db_interface=None):
    """
    Calls protosearch to get the hypothetical structures.

    Args:
        formulas ([str]): list of chemical formulas from which
            to generate candidate structures
        source (str): project name in OQMD to be used as source.
            Defaults to ICSD.
        db_interface (DbInterface): interface to OQMD database
            by default uses the one pulled from data.matr.io

    Returns:
        (pandas.DataFrame) hypothetical pymatgen structures
            generated and their unique ids from protosearch

    TODO:
        - For efficiency, n_max_atoms can be handled within OqmdInterface
    """

    if db_interface is None:
        cache_matrio_data("oqmd_ver3.db")
        oqmd_db_path = os.path.join(CAMD_CACHE, "oqmd_ver3.db")
        db_interface = OqmdInterface(oqmd_db_path)
    dataframes = [
        db_interface.create_proto_data_set(
            source=source, chemical_formula=formula)
        for formula in formulas
    ]
    _structures = pd.concat(dataframes)

    # Drop bad structures
    _structures.dropna(axis=0, how='any', inplace=True)

    # conversion to pymatgen structures
    ase_adap = AseAtomsAdaptor()
    pmg_structures = [ase_adap.get_structure(_structures.iloc[i]['atoms'])
                      for i in range(len(_structures))]
    _structures['pmg_structures'] = pmg_structures

    # The uuid below is probably an overkill. But want the strings
    # to be unique. Sometimes in spaces with similar stoichiometries
    # they may clash e.g. IrSb2O2 and Ir2SbO2 may end up producing
    # the same string, despite different substitutions on same structure.
    # We should figure out a way to get the right order from protosearch.
    structure_uids = [_structures.iloc[i]['proto_name'].replace('_', '-') +
                      '-' + '-'.join(pmg_structures[i].symbol_set) + '-' +
                      str(uuid.uuid4()).replace('-', '')[:6]
                      for i in range(len(_structures))]
    _structures.index = structure_uids
    return _structures


def get_stoichiometric_formulas(n_components, grid=None):
    """
    Args:
        n_components (int): number of components (dimensions)
        grid (list): a range of integers

    Returns:
        (list): unique stoichiometric formula from an
            allowed grid of integers.
    """
    grid = grid if grid else list(range(1, 8))
    args = [grid for _ in range(n_components)]
    stoics = np.array(list(itertools.product(*args)))
    fracs = stoics.astype(float)/np.sum(stoics, axis=1)[:, None]
    _, indices, counts = np.unique(fracs, axis=0, return_index=True,
                                   return_counts=True)
    return stoics[indices]


def create_formulas(bounds, charge_balanced=True, oxi_states_extend=None,
                    oxi_states_override=None, all_oxi_states=False,
                    create_subsystems=False, grid=None):
    """
    Creates a list of formulas given the bounds of a chemical space.
    TODO:
        - implement create_subsystems
    """
    stoichs = get_stoichiometric_formulas(len(bounds), grid=grid)

    formulas = []
    for f in stoichs:
        f_ = ''
        for i in range(len(f)):
            f_ += bounds[i] + f.astype(str).tolist()[i]
        formulas.append(f_)

    if charge_balanced:

        charge_balanced_formulas = []

        if oxi_states_extend:
            oxi_states_override = oxi_states_override if oxi_states_override else {}
            for k, v in oxi_states_extend.items():
                v = v if type(v) == list else [v]
                _states = v + list(Element[k].common_oxidation_states)
                if k in oxi_states_override:
                    oxi_states_override[k] += v
                else:
                    oxi_states_override[k] = _states

        for formula in formulas:
            c = Composition(formula)
            if c.oxi_state_guesses(oxi_states_override=oxi_states_override, all_oxi_states=all_oxi_states):
                charge_balanced_formulas.append(formula)
        return charge_balanced_formulas
    else:
        return formulas


def heuristic_setup(elements):
    """
    Helper function to setup a default structure_domain
    """
    grid_defaults = {2:5, 3:5}
    Ncomp = len(elements)
    _g = grid_defaults.get(Ncomp, 4)
    if {"O", "Cl", "F", "S", "N", "Br", "I"}.intersection(set(elements)):
        charge_balanced = True
    else:
        charge_balanced = False
    if not charge_balanced:
        return _g, charge_balanced
    else:
        g_max_max = 8
        while True:
            sd = StructureDomain.from_bounds(elements, charge_balanced=True,
                                             grid=range(1, _g))
            N = len(sd.formulas)
            if N >= 20:
                return _g, charge_balanced
            else:
                if _g < g_max_max:
                    _g += 1
                else:
                    return _g, charge_balanced



import unittest

import pandas as pd
import numpy as np
from camd.experiment.base import ATFSampler


class ATFSamplerTest(unittest.TestCase):
    def setUp(self):
        test_dataframe = pd.DataFrame({"index": np.arange(5),
                                       "squared": np.arange(5) ** 2})
        params = {"dataframe": test_dataframe,
                  "index_values": [0, 2, 3]}
        self.simple_exp = ATFSampler(params)

    def test_construction(self):
        test_dataframe = pd.DataFrame({"index": np.arange(5),
                                       "squared": np.arange(5) ** 2})
        params = {"dataframe": test_dataframe}
        simple_exp = ATFSampler(params)
        self.assertEqual(simple_exp.get_state(), "completed")
        self.assertTrue(
            (simple_exp.get_results([0, 2, 3])['squared'] == np.array([0, 4, 9])).all())


if __name__ == '__main__':
    unittest.main()



# Copyright 2019, Toyota Research Institute
"""
Logging configuration for operation of CAMD

CAMD logging can be configured by two environment variables:

CAMD_LOG_CONFIG (str): a string that defines the logging handlers
    to be added.  If the string contains "kinesis", the kinesis
    handler is added to the logger, if "stdout", the log is printed
    to stdout.

    Examples:
        "kinesis" - logs to kinesis, kinesis log name can be specified
            via CAMD_KINESIS_STREAM_NAME below
        "stdout" - logs only to stdout
        "kinesis,stdout" - logs to both kinesis and stdout

CAMD_KINESIS_STREAM_NAME (str): stream name for CAMD kinesis
    events.  Defaults to "kinesis-test" but can be configured,
    e. g. to camd-events for production runs

"""
import logging, sys, os
from autologging import TRACE, traced
import boto3


class KinesisHandler(logging.Handler):
    """
    Simple handler for kinesis logging
    """

    def __init__(self, level, stream_name, **kwargs):
        super(KinesisHandler, self).__init__(level=level)
        self.stream_name = stream_name
        self.client = boto3.client('kinesis', **kwargs)

    def emit(self, record):
        try:
            self.client.put_record(StreamName=self.stream_name,
                                   Data=self.format(record),
                                   PartitionKey=str(hash('test'))
                                   )
        except:
            self.handleError(record)


CAMD_LOG_FORMAT = "%(levelname)s:%(name)s:%(funcName)s:%(message)s"
CAMD_LOGGER = logging.getLogger('root')
CAMD_LOGGER.setLevel(TRACE)
CAMD_LOG_FORMATTER = logging.Formatter(CAMD_LOG_FORMAT)

STREAM_HANDLER = logging.StreamHandler(sys.stdout)
STREAM_HANDLER.setLevel(TRACE)
STREAM_HANDLER.setFormatter(CAMD_LOG_FORMATTER)

KINESIS_STREAM_NAME = os.environ.get("CAMD_KINESIS_STREAM_NAME", "kinesis-test")
KINESIS_HANDLER = KinesisHandler(TRACE, "kinesis-test",
                                 region_name="us-west-2")
KINESIS_HANDLER.setFormatter(CAMD_LOG_FORMATTER)

CAMD_LOG_CONFIG = os.environ.get("CAMD_LOG_CONFIG", "")

if "kinesis" in CAMD_LOG_CONFIG:
    CAMD_LOGGER.addHandler(KINESIS_HANDLER)

if "stdout" in CAMD_LOG_CONFIG:
    CAMD_LOGGER.addHandler(STREAM_HANDLER)


def camd_traced(obj):
    """Custom decorator that ensures tracing decorator uses camd logger"""
    return traced(CAMD_LOGGER)(obj)



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
Module and script for running the CAMD campaign worker

Usage:
    camd_worker COMMAND [options]

Options:
    --campaign CAMPAIGN  campaign name  [default: proto-dft-2]
    --loops NUM_LOOPS    number of loops to run
    -h --help            Show this screen
    --version            Show version

COMMAND may be one of:
    start - starts worker
    stop - writes stopfile such that CAMD stops
"""
import time
import boto3
import itertools
import os
import numpy as np

from pathlib import Path
from docopt import docopt
from monty.tempfile import ScratchDir
from camd import CAMD_S3_BUCKET, CAMD_STOP_FILE
from camd.log import camd_traced
from camd.campaigns.structure_discovery import run_proto_dft_campaign, run_atf_campaign


@camd_traced
class Worker(object):
    def __init__(self, campaign="proto-dft-2"):
        self.campaign = campaign

    def start(self, num_loops=np.inf, sleep_time=60):
        """
        Starts the worker, which monitors s3 for new submissions
        and starts runs accordingly.

        Args:
            num_loops (int): number of campaigns to run before
                stopping
            sleep_time (float): time to sleep between iterations
                in which active campaigns are found

        Returns:
            (int) number of loops executed, including "sleep" loops,
                where no campaign was executed

        """
        for loop_num in itertools.count(1):
            if loop_num > num_loops or self.check_stop_file():
                return loop_num - 1
            latest_chemsys = self.get_latest_chemsys()
            if latest_chemsys:
                with ScratchDir('.') as sd:
                    print("Running {} in {}".format(latest_chemsys, sd))
                    self.run_campaign(latest_chemsys)
            else:
                print("No new campaigns submitted, sleeping for {} seconds".format(
                    sleep_time))
                time.sleep(sleep_time)

    def run_campaign(self, chemsys):
        """
        Runs the campaign for a given chemsys

        Args:
            chemsys ([str]):

        Returns:
            None

        """
        if self.campaign == "proto-dft-2":
            run_proto_dft_campaign(chemsys)
        elif self.campaign == "oqmd-atf":
            # This is more or less just a test
            run_atf_campaign(chemsys)
        else:
            raise ValueError("Campaign {} is not valid".format(self.campaign))

    def get_latest_chemsys(self):
        """
        Gets the last submitted chemsys

        Returns:
            (str): corresponding to last submitted chemsys

        """
        bucket = boto3.resource("s3").Bucket(CAMD_S3_BUCKET)

        # Get submissions
        submission_prefix = '/'.join([self.campaign, "submit"])
        submit_objects = bucket.objects.filter(Prefix=submission_prefix)
        submission_times = {obj.key.split('/')[-2]: obj.get()['LastModified']
                            for obj in submit_objects
                            if obj.key != "{}/submit/".format(self.campaign)}

        # Get started jobs
        started = get_common_prefixes(CAMD_S3_BUCKET, "/".join([self.campaign, "runs"]))

        # Filter started jobs and then get latest unstarted
        unstarted = list(set(submission_times.keys()) - set(started))
        latest_unstarted = sorted(unstarted, key=lambda x: submission_times[x])

        return latest_unstarted[-1] if latest_unstarted else None

    @staticmethod
    def write_stop_file():
        """
        Touches stop file to signal drawdown to shared workers

        Returns:
            (None)

        """
        Path(CAMD_STOP_FILE).touch()

    @staticmethod
    def remove_stop_file():
        """
        Removes the stop file from its specified location
        via CAMD_STOP_FILE env variable

        Returns:
            (None)

        """
        if os.path.isfile(CAMD_STOP_FILE):
            os.remove(CAMD_STOP_FILE)

    @staticmethod
    def check_stop_file():
        """
        Returns bool corresponding to whether the CAMD stop
        file exists

        Returns:
            (bool) whether CAMD stop file exists

        """
        return os.path.isfile(CAMD_STOP_FILE)


# TODO: move this to TRI-utils
def get_common_prefixes(bucket, prefix):
    """
    Helper function to get common "subfolders" of folders
    in S3

    Args:
        bucket (str): bucket name
        prefix (str): prefix for which to list common prefixes

    Returns:

    """
    if not prefix.endswith('/'):
        prefix += "/"
    client = boto3.client('s3')
    paginator = client.get_paginator('list_objects')
    result = paginator.paginate(Bucket=bucket, Delimiter='/', Prefix=prefix)
    return [common_prefix['Prefix'].split('/')[-2]
            for common_prefix in result.search("CommonPrefixes")
            if common_prefix]


def main():
    """
    Main function for running this as a script

    Returns:
        None
    """
    args = docopt(__doc__)
    campaign = args['--campaign']
    worker = Worker(campaign)
    num_loops = args['--loops']
    if num_loops is not None:
        num_loops = int(num_loops)
    else:
        num_loops = np.inf
    if args['COMMAND'] == "start":
        print("Starting {} worker with {} loops".format(
            campaign, num_loops))
        worker.remove_stop_file()
        worker.start(num_loops=num_loops)
    elif args['COMMAND'] == "stop":
        worker.write_stopfile()
    else:
        raise ValueError("Invalid command {}.  Worker command must be 'start' or 'stop'")


if __name__ == "__main__":
    main()



# Copyright Toyota Research Institute 2019

import abc

from monty.json import MSONable
from camd.log import camd_traced


class Experiment(abc.ABC, MSONable):
    """
    An abstract class for brokering experiments, this is a
    temporary placeholder for the way we do experiments

    Eventually, this step should execute asynchronously,
    we might just want to use FireWorks for this
    """

    def __init__(self, params):
        self._params = params
        self.unique_ids = params['unique_ids'] if 'unique_ids' in params else []
        self.job_status = params['job_status'] if 'job_status' in params else {}

    @abc.abstractmethod
    def get_state(self):
        """
        # TODO: refine this into something more universal
        Returns:
            str: 'unstarted', 'pending', 'completed'

        """

    @abc.abstractmethod
    def get_results(self, indices):
        """
        Args:
            indices (list): uids / indices of experiments to get the results for
        Returns:
            dict: a dictionary of results

        """

    def get_parameter(self, parameter_name):
        """
        Args:
            parameter_name (str): name of parameter to get

        Returns:
            parameter value

        """
        return self._params[parameter_name]

    def _update_results(self, indices):
        self.results = self.get_results(indices)

    @abc.abstractmethod
    def monitor(self):
        """
        Keeps track of jobs given the poll_time and timeout
        """

    @abc.abstractmethod
    def submit(self, unique_ids):
        """
        # Accepts job requests by unique id of candidates
        Returns:
            str: 'unstarted', 'pending', 'completed'

        """

    @classmethod
    def from_job_status(cls, params, job_status):
        params["job_status"] = job_status
        params["unique_ids"] = list(job_status.keys())
        return cls(params)


@camd_traced
class ATFSampler(Experiment):
    """
    A simple after the fact sampler that just samples
    a dataframe according to index_values
    """

    def start(self):
        """There's no start procedure for this particular experiment"""
        pass

    def get_state(self):
        """This experiment should be complete on construction"""
        return "completed"

    def get_results(self, index_values):
        dataframe = self.get_parameter('dataframe')
        return dataframe.loc[index_values].dropna(axis=0, how='any')

    def submit(self, index_values):
        """This does nothing, since the "experiments" are already done"""
        return {index_value: "completed" for index_value in index_values}

    def monitor(self):
        return True



# Copyright Toyota Research Institute 2019
import os
import pickle
import json
import time
import numpy as np
import pandas as pd
import shutil

from monty.json import MSONable
from camd.utils.data import load_dataframe, s3_sync
from camd import CAMD_S3_BUCKET
from camd.agent.base import RandomAgent
from camd.log import camd_traced
from pymatgen.util.plotting import pretty_plot


@camd_traced
class Loop(MSONable):
    def __init__(self, candidate_data, agent, experiment, analyzer,
                 agent_params=None, experiment_params=None, analyzer_params=None,
                 finalizer=None, finalizer_params=None, path=None, seed_data=None,
                 create_seed=False, heuristic_stopper=False, s3_prefix=None,
                 s3_bucket=CAMD_S3_BUCKET):
        """
        Loop provides a sequential, workflow-like capability where an Agent iterates over a candidate
        space to choose and execute new Experiments, given a certain objective. The abstraction
        follows closely the "scientific method". Agent is the entity that suggests new Experiments.
        Supporting entities are Analyzers and Finalizers. Framework is flexible enough to implement
        many sequential learning or optimization tasks, including active-learning, bayesian optimization
        or black-box optimization with local or global optima search.

        Args:
            candidate_data (pd.DataFrame): List of uids for candidate
                search space for active learning
            agent (HypothesisAgent): a subclass of HypothesisAgent
            experiment (Experiment): a subclass of Experiment
            analyzer (Analyzer): a subclass of Analyzer
            agent_params (dict): parameters of the agent
            experiment_params (dict): parameters of the experiment
            analyzer_params (dict): parameters of the analyzer
            finalizer (obj): needs to have a finalize method.
            finalizer_params (dict): parameters of finalizer
            path (str): path in which to execute the loop. Defaults
                to full path of current folder if not given.
            seed_data (pandas.DataFrame): Seed Data for active learning,
                index is to be the assumed uid
            create_seed (int): an initial seed size to create from the data
            heuristic_stopper (int or False): If int, the heuristic stopper
                will kick in to check if loop should be terminated after
                this many iterations, if no discoveries in past #n loops.
            s3_prefix (str): prefix which to prepend all s3 synced files with,
                if None is specified, s3 syncing will not occur
            s3_bucket (str): bucket name for s3 sync.  If not specified,
                CAMD will sync to the specified environment variable.
        """
        self.path = path if path else '.'
        self.path = os.path.abspath(self.path)
        os.chdir(self.path)

        self.s3_prefix = s3_prefix
        self.s3_bucket = s3_bucket

        self.candidate_data = candidate_data
        self.candidate_space = list(candidate_data.index)

        self.agent = agent(**agent_params)
        self.agent_params = agent_params

        self.experiment = experiment(experiment_params)
        self.experiment_params = experiment_params

        self.analyzer = analyzer(**analyzer_params)
        self.analyzer_params = analyzer_params

        self.finalizer_params = finalizer_params if finalizer_params else {}
        self.finalizer = finalizer(**self.finalizer_params) if finalizer else None

        self.seed_data = seed_data if seed_data is not None else pd.DataFrame()
        self.create_seed = create_seed

        self.heuristic_stopper = heuristic_stopper
        self._exp_raw_results = None
        self._discovered = None
        self.results_all_uids = None
        self.results_new_uids = None

        # Check if there exists earlier iterations
        if os.path.exists(os.path.join(self.path, 'iteration.json')):
            self.load('iteration')
            self.initialized = True
        else:
            self.iteration = 0
            self.initialized = False

        if self.initialized:
            self.create_seed = False
            self.load('job_status')
            self.experiment = self.experiment.from_job_status(
                self.experiment_params, self.job_status)
            self.load('submitted_experiment_requests')
            self.load('seed_data', method='pickle')
            self.load('consumed_candidates')
            self.load('loop_state', no_exist_fail=False)
            self.initialized = True
        else:
            self.submitted_experiment_requests = []
            self.consumed_candidates = []
            self.job_status = {}
            self.initialized = False
            self.loop_state = "UNSTARTED"

    def run(self, finalize=False):
        """
        This method applies a single iteration of the loop,
        and keeps record of everything in place.

        Each iteration consists of:
            1. Get results of requested experiments
            2. Load, Expand, Save seed_data
            3. Augment candidate_space
            4. Analyze results - Stop / Go
            5. Hypothesize
            6. Submit new experiments
        """
        if not self.initialized:
            raise ValueError("Loop needs to be properly initialized.")

        # Get new results
        print("Loop {} state: Getting new results".format(self.iteration))
        self.load('submitted_experiment_requests')
        new_experimental_results = self.experiment.get_results(
            self.submitted_experiment_requests)
        os.chdir(self.path)

        # Load, expand, save seed_data
        self.load('seed_data', method='pickle')
        self.seed_data = self.seed_data.append(new_experimental_results)
        self.save('seed_data', method='pickle')

        # Augment candidate space
        self.load("consumed_candidates")
        self.candidate_space = self.candidate_data.index.difference(
            self.consumed_candidates, sort=False).tolist()
        self.candidate_data = self.candidate_data.loc[self.candidate_space]

        # Analyze results
        print("Loop {} state: Analyzing results".format(self.iteration))
        self.results_new_uids, self.results_all_uids = self.analyzer.analyze(
            self.seed_data, self.submitted_experiment_requests, self.consumed_candidates)
        self.analyzer.present(
            self.seed_data, self.submitted_experiment_requests, self.consumed_candidates,
            filename="hull_{}.png".format(self.iteration))

        self._discovered = np.array(self.submitted_experiment_requests)[self.results_new_uids].tolist()
        self.save('_discovered', custom_name='discovered_{}.json'.format(self.iteration))

        self.report()

        # Loop stopper if no discoveries in last few cycles.
        if self.heuristic_stopper and self.iteration > self.heuristic_stopper and \
                np.sum(pd.read_csv(os.path.join(self.path, 'report.log'), delimiter='\s+')['N_Discovery'][-3:].values) == 0:
            self.finalize()
            raise ValueError("Not enough new discoveries. Stopping the loop.")

        # Loop stopper if finalization is desired but will be done outside of run (e.g. auto_loop)
        if finalize:
            return None

        # Agent suggests new experiments
        print("Loop {} state: Agent {} hypothesizing".format(
            self.iteration, self.agent.__class__.__name__))
        suggested_experiments = self.agent.get_hypotheses(
            self.candidate_data, self.seed_data)

        # Loop stopper if agent doesn't have anything to suggest.
        if len(suggested_experiments) == 0:
            self.finalize()
            raise ValueError("No space left to explore. Stopping the loop.")

        # Experiments submitted
        print("Loop {} state: Running experiments".format(self.iteration))
        self.job_status = self.experiment.submit(suggested_experiments)
        self.save("job_status")

        self.submitted_experiment_requests = suggested_experiments
        self.save('submitted_experiment_requests')

        self.consumed_candidates += suggested_experiments
        self.save('consumed_candidates')

        self.iteration += 1
        self.save("iteration")

    def auto_loop(self, n_iterations=10, timeout=10, monitor=False,
                  initialize=False, with_icsd=False):
        """
        Runs the loop repeatedly, and locally. Pretty light weight, i
        but recommended method is auto_loop_in_directories.
        TODO: Stopping criterion from Analyzer

        Args:
            n_iterations (int): Number of iterations.
            timeout (int): Time (in seconds) to wait on idle for
                submitted experiments to finish.
            monitor (bool): Use Experiment's monitor method to
                keep track of requested experiments.
            initialize (bool): whether to initialize the loop
                before starting
            with_icsd (bool): whether to initialize from icsd

        """
        if initialize:
            if with_icsd:
                self.initialize_with_icsd_seed()
            else:
                self.initialize()
            time.sleep(timeout)
        while n_iterations - self.iteration >= 0:
            print("Iteration: {}".format(self.iteration))
            self.run()
            print("  Waiting for next round ...")
            if monitor:
                self.experiment.monitor()
            time.sleep(timeout)
        self.run(finalize=True)
        self.finalize()

    def auto_loop_in_directories(self, n_iterations=10, timeout=10,
                                 monitor=False, initialize=False,
                                 with_icsd=False):
        """
        Runs the loop repeatedly in directories for each iteration
        TODO: Stopping criterion from Analyzer

        Args:
            n_iterations (int): Number of iterations.
            timeout (int): Time (in seconds) to wait on idle for
                submitted experiments to finish.
            monitor (bool): Use Experiment's monitor method to keep
                track of requested experiments. Note, if this is set
                True, timeout also needs to be adjusted. If this is
                not set True, make sure timeout is sufficiently long.
            initialize (bool): whether to initialize the loop
            with_icsd (bool): whether to initialize with icsd seed

        """
        if initialize:
            self.loop_state = 'AGENT'
            self.save("loop_state")

            if with_icsd:
                self.initialize_with_icsd_seed()
            else:
                self.initialize()

            self.loop_state = 'EXPERIMENTS STARTED'
            self.save("loop_state")

            if monitor:
                self.experiment.monitor()
                os.chdir(self.path)
            time.sleep(timeout)

            if self.experiment.get_state():
                self._exp_raw_results = self.experiment.job_status
                self.save('_exp_raw_results')

            loop_backup(self.path, '-1')
            self.loop_state = 'EXPERIMENTS COMPLETED'
            self.save("loop_state")

        while n_iterations - self.iteration >= 0:
            self.load("loop_state")

            if self.loop_state in ["EXPERIMENTS COMPLETED", "AGENT"]:
                print("Iteration: {}".format(self.iteration))
                self.loop_state = 'AGENT'
                self.save("loop_state")
                self.run()
                print("  Waiting for next round ...")

            self.loop_state = 'EXPERIMENTS STARTED'
            self.save("loop_state")
            if monitor:
                self.experiment.monitor()
                os.chdir(self.path)
            time.sleep(timeout)
            if self.experiment.get_state():
                self._exp_raw_results = self.experiment.job_status
                self.save('_exp_raw_results')

            loop_backup(self.path, str(self.iteration - 1))
            self.loop_state = 'EXPERIMENTS COMPLETED'
            self.save("loop_state")
        self.run(finalize=True)
        self.finalize()

    def initialize(self, random_state=42):
        if self.initialized:
            raise ValueError("Initialization may overwrite existing loop data. Exit.")
        if not self.seed_data.empty and not self.create_seed:
            print("Loop {} state: Agent {} hypothesizing".format(
                'initialization', self.agent.__class__.__name__))
            suggested_experiments = self.agent.get_hypotheses(
                self.candidate_data, self.seed_data)
        elif self.create_seed:
            np.random.seed(seed=random_state)
            _agent = RandomAgent(self.candidate_data, n_query=self.create_seed)
            print("Loop {} state: Agent {} hypothesizing".format(
                'initialization', _agent.__class__.__name__))
            suggested_experiments = _agent.get_hypotheses(self.candidate_data)
        else:
            raise ValueError("No seed data available. "
                             "Either supply or ask for creation.")

        if len(suggested_experiments) == 0:
            raise ValueError("No hypotheses found to initialize with. Stopping the loop.")

        print("Loop {} state: Running experiments".format(self.iteration))
        self.job_status = self.experiment.submit(suggested_experiments)

        self.submitted_experiment_requests = suggested_experiments
        self.consumed_candidates = suggested_experiments
        self.create_seed = False
        self.initialized = True

        self.save("job_status")
        self.save("seed_data", method='pickle')
        self.save('submitted_experiment_requests')
        self.save("consumed_candidates")
        self.save("iteration")
        if self.s3_prefix:
            self.s3_sync()

    def initialize_with_icsd_seed(self, random_state=42):
        if self.initialized:
            raise ValueError("Initialization may overwrite existing loop data. Exit.")
        self.seed_data = load_dataframe("oqmd1.2_icsd_featurized_clean_v2.pickle")
        self.initialize(random_state=random_state)

    def report(self):
        with open(os.path.join(self.path, 'report.log'), 'a') as f:
            if self.iteration == 0:
                f.write("Iteration N_Discovery Total_Discovery N_candidates model-CV\n")
            report_string = "{:9} {:11} {:15} {:12} {:f}\n".format(
                self.iteration, np.sum(self.results_new_uids),
                np.sum(self.results_all_uids), len(self.candidate_data),
                self.agent.cv_score)
            f.write(report_string)

        self.generate_report_plot()

    def finalize(self):
        print("Finalizing campaign.")
        os.chdir(self.path)
        if self.finalizer:
            self.finalizer.finalize(self.path)
        if self.s3_prefix:
            self.s3_sync()

    @staticmethod
    def generate_report_plot(filename="report.png",
                             report_filename="report.log"):
        """
        Quick method for generating report plots

        Args:
            filename (str): output filename for plot to be saved
            report_filename (str): filename for the report to be read in

        Returns:
            (AxesSubplot) pyplot object corresponding to bar plot

        """
        # Generate plot
        data = pd.read_csv(report_filename, delim_whitespace=True)
        plt = pretty_plot(6, 4.5)
        ax = plt.gca()
        ax = data.plot(kind='bar', x='Iteration', y='Total_Discovery',
                       legend=False, ax=ax)
        ax.set_ylabel("Total materials discovered")
        fig = ax.get_figure()
        if filename:
            fig.savefig(filename, dpi=70)

        # Close to avoid matplotlib memory warning
        plt.close()

    def load(self, data_holder, method='json', no_exist_fail=True):
        if method == 'pickle':
            m = pickle
            mode = 'rb'
        elif method == 'json':
            m = json
            mode = 'r'
        else:
            raise ValueError("Unknown data save method")

        file_name = os.path.join(self.path, data_holder+'.'+method)
        exists = os.path.exists(file_name)

        if exists:
            with open(file_name, mode) as f:
                self.__setattr__(data_holder, m.load(f))
        else:
            if no_exist_fail:
                raise IOError("No {} file exists".format(data_holder))
            else:
                self.__setattr__(data_holder, None)

    def save(self, data_holder, custom_name=None, method='json'):
        if custom_name:
            _path = os.path.join(self.path, custom_name)
        else:
            _path = os.path.join(self.path, data_holder+'.'+method)
        if method == 'pickle':
            m = pickle
            mode = 'wb'
        elif method == 'json':
            m = json
            mode = 'w'
        else:
            raise ValueError("Unknown data save method")
        with open(_path, mode) as f:
            m.dump(self.__getattribute__(data_holder), f)

        # Do s3 sync if present
        if self.s3_prefix:
            self.s3_sync()

    def get_state(self):
        return self.loop_state

    def s3_sync(self):
        """
        Syncs current run to s3_prefix and bucket
        """
        s3_sync(self.s3_bucket, self.s3_prefix, self.path)


def loop_backup(src, new_dir_name):
    """
    Helper method to backup finished loop iterations.
    Args:
        src:
        new_dir_name:

    Returns:

    """
    os.mkdir(os.path.join(src, new_dir_name))
    _files = os.listdir(src)
    for file_name in _files:
        full_file_name = os.path.join(src, file_name)
        if os.path.isfile(full_file_name):
            shutil.copy(full_file_name, new_dir_name)



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.

import unittest
import os

from pymatgen.util.testing import PymatgenTest
from pymatgen import MPRester
from camd.experiment.dft import OqmdDFTonMC1


CAMD_DFT_TESTS = os.environ.get("CAMD_DFT_TESTS", False)
SKIP_MSG = "Long tests disabled, set CAMD_DFT_TESTS to run long tests"


# This test is still inconsistent because of issues with
# batch AWS jobs and database communications
class Mc1Test(unittest.TestCase):
    @unittest.skipUnless(CAMD_DFT_TESTS, SKIP_MSG)
    def test_get(self):
        good_silicon = PymatgenTest.get_structure("Si")
        bad_silicon = good_silicon.copy()

        # Add another site at the same position
        bad_silicon.append("Si", [0.1, 0.1, 0.15])
        bad_silicon.append("Si", [0.1, 0.333, 0.15])
        self.assertEqual(len(bad_silicon), 4)
        params = {"structure_dict": {"good": good_silicon,
                                     "bad": bad_silicon},
                  "poll_time": 30, "timeout": 300}

        experiment = OqmdDFTonMC1(params)
        experiment.submit()
        status = experiment.monitor()
        results = experiment.get_results(
            ['good', 'bad'], populate_candidate_data=False)

        self.assertAlmostEqual(results['good'], 0, 5)
        self.assertIsNone(results.get('bad'))

    @unittest.skipUnless(CAMD_DFT_TESTS, SKIP_MSG)
    def test_structure_suite(self):
        # TODO: fix the formation energy calculation
        mp_ids = ["mp-702",
                  "mp-1953",
                  "mp-1132",
                  "mp-8409",
                  "mp-872"]
        with MPRester() as mpr:
            structure_dict = {mp_id: mpr.get_structure_by_material_id(mp_id)
                              for mp_id in mp_ids}
        params = {"structure_dict": structure_dict,
                  "poll_time": 25}

        experiment = OqmdDFTonMC1(params)
        experiment.submit()
        status = experiment.monitor()
        results = experiment.get_results(mp_ids)
        self.assertTrue(all([run['status'] == "SUCCEEDED" for run in status]))


if __name__ == '__main__':
    unittest.main()



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
This module consolidates s3-based dataset loading
for CAMD, mostly in the form of dataframes
"""

import os

import boto3
import requests
import pandas as pd
from monty.os import makedirs_p
from camd import CAMD_CACHE, tqdm

# TODO-PUBLIC: this will need to be made general for the
#   release of the code, mostly by making the s3 bucket public


def load_dataframe(dataset_name):
    """
    Helper function to load a dataframe from the utilities below

    Supported datasets include:
        * oqmd_1.2_voronoi_magpie_fingerprints
        * oqmd1.2_exp_based_entries_featurized_v2,

    Args:
        dataset_name (str): dataset name to load

    Returns:
        (DataFrame): dataframe corresponding to specified dataset

    """
    filename = '{}.pickle'.format(dataset_name)
    cache_matrio_data(filename)
    return pd.read_pickle(os.path.join(CAMD_CACHE, filename))


def load_default_atf_data():
    """
    Convenience function to load the default test ATF data
    which is a dataframe of OQMD data of binary compounds with
    magpie and voronoi analysis features

    Returns:
        (DataFrame): dataframe of OQMD data and the corresponding
            features

    """
    df = load_dataframe("oqmd_1.2_voronoi_magpie_fingerprints")
    return df[df['N_species'] == 2].sample(frac=0.2)


def cache_download(url, path):
    """
    Quick helper function to cache a generic download from a url
    in the CAMD local data directory

    Args:
        url (str): url for download
        path (str): path for download, is appended to the
            CAMD_CACHE location

    Returns:
        (None)
    """
    # Prep cache path and make necessary dirs
    cache_path = os.path.join(CAMD_CACHE, path)

    # Download and write file
    if not os.path.isfile(cache_path):
        makedirs_p(os.path.split(cache_path)[0])
        r = requests.get(url, stream=True)
        total_size = int(r.headers.get('content-length', 0))
        block_size = 1024  # 1 Kibibyte
        t = tqdm(total=total_size, unit='iB', unit_scale=True)
        with open(cache_path, 'wb') as f:
            for data in r.iter_content(block_size):
                t.update(len(data))
                f.write(data)


# Mapping of matrio data hashes to cache paths
MATRIO_DATA_KEYS = {
    "oqmd1.2_exp_based_entries_featurized_v2.pickle": "5e3b0e9bc91e209071f33ce8",
    "oqmd_1.2_voronoi_magpie_fingerprints.pickle": "5e39ce2cd9f13e075b7dfaaf",
    "oqmd_ver3.db": "5e39ce96d9f13e075b7dfab3",
    "oqmd1.2_exp_based_entries_structures.json": "5e45befef23b399192b35242"
}


def cache_matrio_data(filename):
    """
    Helper function to cache data hosted on data.matr.io

    Args:
        filename (filename): filename to fetch from data.matr.io

    Returns:
        None

    """
    prefix = "https://data.matr.io/3/api/v1/file"
    key = MATRIO_DATA_KEYS[filename]
    if not os.path.isfile(filename):
        cache_download("{}/{}/download".format(prefix, key), filename)


def s3_sync(s3_bucket, s3_prefix, sync_path="."):
    """
    Syncs a given path to an s3 prefix

    Args:
        s3_bucket (str): bucket name
        s3_prefix (str): s3 prefix to sync to
        sync_path (str, Path): path to sync to bucket:prefix

    Returns:
        (None)

    """
    # Get bucket
    s3_resource = boto3.resource("s3")
    bucket = s3_resource.Bucket(s3_bucket)

    # Walk paths and subdirectories, uploading files
    for path, subdirs, files in os.walk(sync_path):
        # Get relative path prefix
        relpath = os.path.relpath(path, sync_path)
        if not relpath.startswith('.'):
            prefix = os.path.join(s3_prefix, relpath)
        else:
            prefix = s3_prefix

        for file in files:
            file_key = os.path.join(prefix, file)
            bucket.upload_file(os.path.join(path, file), file_key)


#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.

import traceback
import pandas as pd
import os

from datetime import datetime
from monty.serialization import dumpfn
from camd.domain import StructureDomain, heuristic_setup
from camd.agent.agents import AgentStabilityAdaBoost
from camd.agent.base import RandomAgent
from camd.experiment import ATFSampler
from camd.loop import Loop
from camd import CAMD_TEST_FILES, CAMD_S3_BUCKET
from camd.utils.data import s3_sync

from camd.analysis import AnalyzeStability, FinalizeQqmdCampaign
from camd.experiment.dft import OqmdDFTonMC1
from sklearn.neural_network import MLPRegressor
import pickle


__version__ = "2019.09.16"


# TODO: abstract campaign?

def run_proto_dft_campaign(chemsys, s3_prefix="proto-dft-2"):
    """

    Args:
        chemsys (str): chemical system for the campaign
        s3_prefix (str): s3 prefix to sync to

    Returns:
        (bool): True if run exits

    """
    s3_prefix = "{}/runs/{}".format(s3_prefix, chemsys)

    # Initialize s3
    dumpfn({"started": datetime.now().isoformat(),
            "version": __version__}, "start.json")
    s3_sync(s3_bucket=CAMD_S3_BUCKET, s3_prefix=s3_prefix, sync_path='.')

    try:
        # Get structure domain
        element_list = chemsys.split('-')
        g_max, charge_balanced = heuristic_setup(element_list)
        domain = StructureDomain.from_bounds(
            element_list, charge_balanced=charge_balanced,
            n_max_atoms=20, **{'grid': range(1, g_max)})
        candidate_data = domain.candidates()
        structure_dict = domain.hypo_structures_dict

        # Dump structure/candidate data
        with open('candidate_data.pickle', 'wb') as f:
            pickle.dump(candidate_data, f)
        with open('structure_dict.pickle', 'wb') as f:
            pickle.dump(structure_dict, f)

        # Set up agents and loop parameters
        agent = AgentStabilityAdaBoost
        agent_params = {
            'ml_algorithm': MLPRegressor,
            'ml_algorithm_params': {'hidden_layer_sizes': (84, 50)},
            'n_query': 10,
            'hull_distance': 0.2,  # Distance to hull to consider a finding as discovery (eV/atom)
            'exploit_fraction': 1.0,  # Fraction to exploit (rest will be explored -- randomly picked)
            'uncertainty': True,
            'alpha': 0.5,
            'diversify': True,
            'n_estimators': 20
        }
        analyzer = AnalyzeStability
        analyzer_params = {'hull_distance': 0.2}  # analysis criterion (need not be exactly same as agent's goal)
        experiment = OqmdDFTonMC1
        experiment_params = {'structure_dict': structure_dict, 'candidate_data': candidate_data, 'timeout': 30000}
        experiment_params.update({'timeout': 30000})
        finalizer = FinalizeQqmdCampaign
        finalizer_params = {'hull_distance': 0.2}
        n_max_iter = n_max_iter_heuristics(len(candidate_data), 10)

        # Construct and start loop
        new_loop = Loop(
            candidate_data, agent, experiment, analyzer, agent_params=agent_params,
            analyzer_params=analyzer_params, experiment_params=experiment_params,
            finalizer=finalizer, finalizer_params=finalizer_params, heuristic_stopper=5,
            s3_prefix=s3_prefix)
        new_loop.auto_loop_in_directories(
            n_iterations=n_max_iter, timeout=10, monitor=True,
            initialize=True, with_icsd=True)
    except Exception as e:
        error_msg = {"error": "{}".format(e),
                     "traceback": traceback.format_exc()}
        dumpfn(error_msg, "error.json")
        dumpfn({"status": "error"}, "job_status.json")
        s3_sync(s3_bucket=CAMD_S3_BUCKET, s3_prefix=s3_prefix, sync_path='.')

    return True


def run_atf_campaign(chemsys):
    """
    A very simple test campaign

    Returns:
        True

    """
    s3_prefix = "oqmd-atf/runs/{}".format(chemsys)
    df = pd.read_csv(os.path.join(CAMD_TEST_FILES, 'test_df.csv'))
    n_seed = 200  # Starting sample size
    n_query = 10  # This many new candidates are "calculated with DFT" (i.e. requested from Oracle -- DFT)
    agent = RandomAgent
    agent_params = {'n_query': n_query}
    analyzer = AnalyzeStability
    analyzer_params = {'hull_distance': 0.05}
    experiment = ATFSampler
    experiment_params = {'dataframe': df}
    candidate_data = df
    new_loop = Loop(candidate_data, agent, experiment, analyzer,
                    agent_params=agent_params, analyzer_params=analyzer_params,
                    experiment_params=experiment_params, create_seed=n_seed,
                    s3_prefix=s3_prefix)

    new_loop.initialize()

    for _ in range(3):
        new_loop.run()

    return True


def n_max_iter_heuristics(n_data, n_query, low_bound=5, up_bound=20):
    """
    Helper method to define maximum number of iterations for a given campaign.
    This is based on the empirical evidence in various systems >90% of stable
    materials are identified when 25% of candidates are tested. We also enforce
    upper and lower bounds of 20 and 5 to avoid edge cases with too many or too
    few calculations to run.

    Args:
        n_data (int): number of data points in candidate space
        n_query (int): number of queries allowed in each iteration
        low_bound (int): lower bound allowed for n_max_iter
        up_bound (int): upper bound allowed for n_max_ite

    Returns:
        maximum number of iterations as integer

    """
    _target = round(n_data*0.25/n_query)
    if _target<low_bound:
        return low_bound
    else:
        return min(_target, up_bound)



# Copyright Toyota Research Institute 2019

import os
import uuid
import json
import re
import time
import shlex
import subprocess
import traceback
import warnings
import datetime

import pandas as pd
from monty.os import cd
from monty.tempfile import ScratchDir
from pymatgen.io.vasp.outputs import Vasprun
from camd.experiment.base import Experiment


class OqmdDFTonMC1(Experiment):
    def __init__(self, params):

        if 'structure_dict' not in params:
            raise ValueError("A dictionary of structures must be "
                             "provided as input as structure_dict")

        self.structure_dict = params['structure_dict']
        self.poll_time = params['poll_time'] if 'poll_time' in params else 60
        self.timeout = params['timeout'] if 'timeout' in params else 7200

        self.unique_ids = params['unique_ids'] if 'unique_ids' in params else []
        self.job_status = params['job_status'] if 'job_status' in params else {}
        super().__init__(params)

    @property
    def state(self):
        if self.get_state() and self.job_status:
            return "completed"
        elif sum([doc['status'] in ['SUCCEEDED', 'FAILED']
                  for doc in self.job_status.values()]) > 0:
            return "active"
        else:
            return "unstarted"

    @property
    def state_detailed(self):
        return NotImplementedError("Not implemented")

    def get_state(self):
        self.job_status = check_dft_calcs(self.job_status)
        return all([doc['status'] in ['SUCCEEDED', 'FAILED']
                    for doc in self.job_status.values()])

    def get_results(self, indices, populate_candidate_data=True):
        # This gets the formation energies.
        if not self.get_state():
            warnings.warn("Some calculations did not finish.")
        results_dict = {}
        for structure_id, calc in self.job_status.items():
            if calc['status'] == "SUCCEEDED":
                delta_e = get_qmpy_formation_energy(
                    calc['result']['output']['final_energy_per_atom'],
                    calc['result']['pretty_formula'], 1)

                results_dict[structure_id] = delta_e

        if populate_candidate_data:
            candidate_data = self.get_parameter("candidate_data")
            _df = candidate_data.loc[results_dict.keys()]
            _df['delta_e'] = pd.Series(results_dict)
            _df = _df.reindex(sorted(_df.columns), axis=1)
            return _df
        else:
            return results_dict

    def submit(self, unique_ids=None):
        """
        Args:
            unique_ids (list): Unique ids for structures to run from t
                he structure_dict. If None, all entries in structure_dict
                are submitted.

        Returns:
            (str): job status

        """
        self.unique_ids = unique_ids if unique_ids \
            else list(self.structure_dict.keys())
        self._structures_to_run = dict([(k, v) for k, v in self.structure_dict.items()
                                        if k in self.unique_ids])
        self.job_status = submit_dft_calcs_to_mc1(self._structures_to_run)
        return self.job_status

    def monitor(self):
        """
        Returns:
            (dict): calculation status, including results
        """
        finished = self.get_state()
        with ScratchDir("."):
            while not finished:
                time.sleep(self.poll_time)
                finished = self.get_state()
                for doc in self.job_status.values():
                    doc["elapsed_time"] = time.time() - doc["start_time"]
                status_string = "\n".join(["{}: {} {}".format(
                    key, value["status"], datetime.timedelta(0, value["elapsed_time"]))
                    for key, value in self.job_status.items()])
                print("Calc status:\n{}".format(status_string))
                print("Timeout is set as {}.".format(datetime.timedelta(0, self.timeout)))

                for doc in self.job_status.values():
                    doc["elapsed_time"] = time.time() - doc["start_time"]
                    if doc["elapsed_time"] > self.timeout:
                        if doc['status'] not in ['SUCCEEDED', 'FAILED']:
                            # Update job status to reflect timeout
                            doc.update({"status": "FAILED",
                                        "error": "timeout"})
                            # Kill AWS job
                            kill_cmd = "aws batch terminate-job --region=us-east-1 " \
                                       "--job-id {} --reason camd_timeout".format(
                                doc['jobId'])
                            kill_result = subprocess.check_output(shlex.split(kill_cmd))

        return self.job_status

    @classmethod
    def from_job_status(cls, params, job_status):
        params["job_status"] = job_status
        params["unique_ids"] = list(job_status.keys())
        return cls(params)


def submit_dft_calcs_to_mc1(structure_dict):
    """
    Placeholder function for fetching DFT calculations from MC1
    using oqmd container

    Args:
        structure_dict (dict): dictionary of structures keyed by
            some string identifier
    """
    starting_dir = os.getcwd()
    tri_path = os.environ.get("TRI_PATH")
    if not tri_path:
        raise ValueError("TRI_PATH must be specified as env variable to "
                         "use camd MC1 interface")
    # Create run directory
    uuid_string = str(uuid.uuid4()).replace('-', '')
    parent_dir = os.path.join(tri_path, "model", "oqmdvasp", "3",
                              "u", "camd", "run{}".format(uuid_string))
    if any(['_' in key for key in structure_dict.keys()]):
        raise ValueError("Structure keys cannot contain underscores for "
                         "mc1 compatibility")

    calc_status = {}
    for structure_id, structure in structure_dict.items():
        calc_path = os.path.join(parent_dir, structure_id, "_1")
        os.makedirs(calc_path)
        with cd(calc_path):
            # Write input cif file and python model file
            structure.to(filename="POSCAR")
            with open("model.py", "w") as f:
                f.write(MODEL_TEMPLATE)

            # Submit to mc1
            # TODO: ensure this is checked for failure to submit
            print("Submitting job: {}".format(structure_id))
            try:
                calc = subprocess.check_output([
                    "trisub", "-q", "oqmd_test_queue",
                    "-r", "16000", "-c", "16"])
            except subprocess.CalledProcessError as e:
                print(e.output)

            calc = calc.decode('utf-8')
            calc = re.findall("({.+})", calc, re.DOTALL)[0]
            calc = json.loads(calc)
            calc.update({"path": os.getcwd(),
                         "status": "SUBMITTED",
                         "start_time": time.time()})
            calc_status[structure_id] = calc

    return calc_status


def check_dft_calcs(calc_status):
    """

    Args:
        calc_status (dict):

    Returns:
        updated calc_status dictionary

    """
    for structure_id, calc in calc_status.items():
        if calc['status'] in ['SUCCEEDED', 'FAILED']:
            continue
        path = calc['path']
        print("Checking status of {}: {}".format(path, structure_id))
        aws_cmd = "aws batch describe-jobs --jobs " \
                  "--region=us-east-1 {}".format(calc['jobId'])
        result = subprocess.check_output(shlex.split(aws_cmd))
        result = json.loads(result)
        aws_status = result["jobs"][0]["status"]
        if aws_status == "SUCCEEDED":
            os.chdir(path)
            subprocess.call('trisync')
            os.chdir('simulation')
            try:
                vr = Vasprun('static/vasprun.xml')
                calc.update({
                    "status": "SUCCEEDED",
                    "error": None,
                    "result": vr.as_dict()
                })
            except Exception as e:
                error_doc = {}
                with open('err') as errfile:
                    error_doc.update({"trisub_stderr": errfile.read()})
                error_doc.update({"camd_exception": "{}".format(e),
                                  "camd_traceback": traceback.format_exc()})
                calc.update({
                    "status": "FAILED",
                    "error": error_doc,
                    "result": None
                })
        elif aws_status == "FAILED":
            error_doc = {"aws_fail": result['jobs'][0]['attempts'][-1]['statusReason']}
            calc.update({"status": "FAILED",
                         "error": error_doc
                         })
        else:
            calc.update({"status": aws_status})
    return calc_status


MODEL_TEMPLATE = """
import os

import qmpy
from qmpy.materials.structure import Structure
from qmpy.analysis.vasp.calculation import Calculation
from qmpy import io
import time


# TODO: definitely move this somewhere else, as it's not
#       meant to be imported
def run_oqmd_calculation(poscar_filename):
    starting_structure = io.poscar.read(poscar_filename)

    # Relaxation
    os.mkdir("relax")
    os.chdir("relax")
    calc = Calculation()
    calc.setup(starting_structure, "relaxation")
    os.system("mpirun -n 16 vasp_std")
    # Just in case the mysql server process dies
    # Kids, don't try this at home
    os.system("sudo -u mysql mysqld &")
    relaxed_structure = io.poscar.read("CONTCAR")
    os.chdir('..')

    # Relaxation
    os.mkdir("static")
    os.chdir("static")
    calc = Calculation()
    calc.setup(relaxed_structure, "static")
    os.system("mpirun -n 16 vasp_std")
    os.chdir('..')


if __name__ == '__main__':
    run_oqmd_calculation("POSCAR")
"""


def get_qmpy_formation_energy(total_e, formula, n_atoms):
    mus = {u'Ac': -4.1060035325, u'Ag': -2.8217729525, u'Al': -3.74573946, u'Ar': -0.00636995, u'As': -4.651918435,
           u'Au': -3.26680174, u'B': -6.67796758, u'Ba': -1.92352708, u'Be': -3.75520865, u'Bi': -4.038931855,
           u'Br': -1.31759562258416, u'C': -9.2170759925, u'Ca': -1.977817, u'Cd': -0.90043514, u'Ce': -4.7771708225,
           u'Cl': -1.47561368438088, u'Co': -7.089565, u'Cr': -9.50844998, u'Cs': -0.85462775, u'Cu': -3.7159594,
           u'Dy': -4.60150328333333, u'Er': -4.56334055, u'Eu': -1.8875732, u'F': -1.45692429086889, u'Fe': -8.3078978,
           u'Ga': -3.031846515, u'Gd': -4.6550712925, u'Ge': -4.623692585, u'H': -3.38063384781582, u'He': -0.004303435,
           u'Hf': -9.955368785, u'Hg': -0.358963825033731, u'Ho': -4.57679364666667, u'I': -1.35196205757168,
           u'In': -2.71993876, u'Ir': -8.8549203, u'K': -1.096699335, u'Kr': -0.004058825, u'La': -4.93543556,
           u'Li': -1.89660627, u'Lu': -4.524181525, u'Mg': -1.54251595083333, u'Mn': -9.0269032462069,
           u'Mo': -10.8480839, u'N': -8.11974103465649, u'Na': -1.19920373914835, u'Nb': -10.09391206,
           u'Nd': -4.762916335, u'Ne': -0.02931791, u'Ni': -5.56661952, u'Np': -12.94027372125, u'O': -4.52329546412125,
           u'Os': -11.22597601, u'P': -5.15856496104006, u'Pa': -9.49577589, u'Pb': -3.70396484, u'Pd': -5.17671826,
           u'Pm': -4.7452352875, u'Pr': -4.7748066125, u'Pt': -6.05575959, u'Pu': -14.29838348, u'Rb': -0.9630733,
           u'Re': -12.422818875, u'Rh': -7.26940476, u'Ru': -9.2019888, u'S': -3.83888286598664, u'Sb': -4.117563025,
           u'Sc': -6.328367185, u'Se': -3.48117276, u'Si': -5.424892535, u'Sm': -4.7147675825, u'Sn': -3.9140929231488,
           u'Sr': -1.6829138, u'Ta': -11.85252937, u'Tb': -5.28775675533333, u'Tc': -10.360747885,
           u'Te': -3.14184237666667, u'Th': -7.41301719, u'Ti': -7.69805778621374, u'Tl': -2.359420025,
           u'Tm': -4.47502416, u'U': -11.292348705, u'V': -8.94097896, u'W': -12.96020695, u'Xe': 0.00306349,
           u'Y': -6.464420635, u'Yb': -1.51277545, u'Zn': -1.2660268, u'Zr': -8.54717235}
    hubbard_mus = {u'Co': 2.0736240219357, u'Cr': 2.79591214925926, u'Cu': 1.457571831687, u'Fe': 2.24490453841424,
                   u'Mn': 2.08652912841877, u'Ni': 2.56766185643768, u'Np': 2.77764768949249, u'Pu': 2.2108747749433,
                   u'Th': 1.06653674624248, u'U': 2.57513786752409, u'V': 2.67812162528461}

    from pymatgen import Composition
    c = Composition(formula).fractional_composition
    e = total_e / n_atoms
    for k,v in c.as_dict().items():
        e -= mus[k]*v
        if (k in hubbard_mus) and ('O' in c):
            e -= hubbard_mus[k]*v
    return e



from setuptools import setup, find_packages
import warnings

try:
    import numpy
except ImportError:
    # This is crude, but the best way I can figure to do this
    warnings.warn("Setup requires pre-installation of numpy, run pip "
                  "install numpy before setup.py")

DESCRIPTION = "camd is software designed to support autonomous materials " \
              "research and sequential learning"

LONG_DESCRIPTION = """
camd is software designed to support Computational Autonomy for Materials Discovery
based on ongoing work led by the
[Toyota Research Institute](http://www.tri.global/accelerated-materials-design-and-discovery/).

camd enables the construction of sequential learning pipelines using a set of
abstractions that include
* Agents - decision making entities which select experiments to run from pre-determined candidate sets
* Experiments - experimental procedures which augment candidate data in a way that facilitates further experiment selection
* Analyzers - Post-processing procedures which frame experimental results in the context of candidate or seed datasets

In addition to these abstractions, camd provides a loop construct which executes
the sequence of hypothesize-experiment-analyze by the Agent, Experiment, and Analyzer,
respectively.  Simulations of agent performance can also be conducted using
after the fact sampling of known data.
"""

setup(
    name='camd',
    url="https://github.com/TRI-AMDD/CAMD",
    version="2020.3.24-post0",
    packages=find_packages(),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    setup_requires=["numpy==1.18.2",
                    "Django==2.2.10",
                    "tensorflow==1.15.2",
                    "gpflow==1.5.0"],  # For qmpy depnedency
    install_requires=["numpy==1.18",
                      "python-dateutil==2.8.0",
                      "networkx==2.2",
                      "matplotlib==3.1.1",
                      "qmpy",  # This version is constrained by the source
                      "pandas==0.24.2",
                      "matminer==0.5.5",
                      "autologging",
                      "awscli==1.18.27",
                      "docopt==0.6.2",
                      ],
    extras_require={
        "proto_dft": ["protosearch"],
        "tests": ["pytest",
                  "pytest-cov",
                  "coveralls"
                  ]
    },
    dependency_links=[
        "https://github.com/JosephMontoya-TRI/qmpy_py3/tarball/master#egg=qmpy",
        "https://github.com/ToyotaResearchInstitute/protosearch/tarball/master#egg=protosearch",
    ],
    entry_points={
        "console_scripts": [
            "camd_worker = camd.campaigns.worker:main"
        ]
    },
    classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
    ],
    include_package_data=True,
    author="AMDD - Toyota Research Institute",
    author_email="murat.aykol@tri.global",
    maintainer="Murat Aykol, Joseph Montoya",
    maintainer_email="murat.aykol@tri.global",
    license="Apache",
    keywords=[
        "materials", "battery", "chemistry", "science",
        "density functional theory", "energy", "AI", "artificial intelligence",
        "sequential learning", "active learning"
    ],
    )


# Copyright Toyota Research Institute 2019

import abc
import warnings
import json
import pickle
import os
import numpy as np
import itertools
from camd import tqdm
from camd.log import camd_traced
from qmpy.analysis.thermodynamics.phase import Phase, PhaseData
from qmpy.analysis.thermodynamics.space import PhaseSpace
import multiprocessing
from pymatgen import Composition
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.analysis.phase_diagram import PhaseDiagram, PDPlotter, tet_coord,\
    triangular_coord
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen import Structure, Element
from camd.utils.data import cache_matrio_data
from camd import CAMD_CACHE
from monty.os import cd
from monty.serialization import loadfn

ELEMENTS = ['Ru', 'Re', 'Rb', 'Rh', 'Be', 'Ba', 'Bi', 'Br', 'H', 'P',
            'Os', 'Ge', 'Gd', 'Ga', 'Pr', 'Pt', 'Pu', 'C', 'Pb', 'Pa',
            'Pd', 'Xe', 'Pm', 'Ho', 'Hf', 'Hg', 'He', 'Mg', 'K', 'Mn',
            'O', 'S', 'W', 'Zn', 'Eu', 'Zr', 'Er', 'Ni', 'Na', 'Nb',
            'Nd', 'Ne', 'Np', 'Fe', 'B', 'F', 'Sr', 'N', 'Kr', 'Si',
            'Sn', 'Sm', 'V', 'Sc', 'Sb', 'Se', 'Co', 'Cl', 'Ca', 'Ce',
            'Cd', 'Tm', 'Cs', 'Cr', 'Cu', 'La', 'Li', 'Tl', 'Lu', 'Th',
            'Ti', 'Te', 'Tb', 'Tc', 'Ta', 'Yb', 'Dy', 'I', 'U', 'Y', 'Ac',
            'Ag', 'Ir', 'Al', 'As', 'Ar', 'Au', 'In', 'Mo']


class AnalyzerBase(abc.ABC):
    @abc.abstractmethod
    def analyze(self):
        """
        Performs the analysis procedure associated with the analyzer

        Returns:
            Some arbitrary result

        """

    @abc.abstractmethod
    def present(self):
        """
        Formats the analysis into a some presentation-oriented
        document

        Returns:
            json document for presentation, e. g. on a web frontend

        """


@camd_traced
class AnalyzeStructures(AnalyzerBase):
    """
    This class tests if a list of structures are unique. Typically
    used for comparing hypothetical structures (post-DFT relaxation)
    and those from ICSD.
    """
    def __init__(self, structures=None, hull_distance=None):
        self.structures = structures if structures else []
        self.structure_ids = None
        self.unique_structures = None
        self.groups = None
        self.energies = None
        self.against_icsd = False
        self.structure_is_unique = None
        self.hull_distance = hull_distance
        super(AnalyzeStructures, self).__init__()

    def analyze(self, structures=None, structure_ids=None,
                against_icsd=False, energies=None):
        """
        One encounter of a given structure will be labeled as True, its
        remaining matching structures as False.

        Args:
            structures (list): a list of structures to be compared.
            structure_ids (list): uids of structures, optional.
            against_icsd (bool): whether a comparison to icsd is also made.
            energies (list): list of energies (per atom) corresponding
                to structures. If given, the lowest energy instance of a
                given structure will be return as the unique one. Otherwise,
                there is no such guarantee. (optional)
        Returns:
            ([bool]) list of bools corresponding to the given list of
                structures corresponding to uniqueness
        """
        self.structures = structures
        self.structure_ids = structure_ids
        self.against_icsd = against_icsd
        self.energies = energies

        smatch = StructureMatcher()
        self.groups = smatch.group_structures(structures)
        self.structure_is_unique = []

        if self.energies:
            for i in range(len(self.groups)):
                self.groups[i] = [x for _, x in sorted(
                    zip([self.energies[self.structures.index(s)]
                         for s in self.groups[i]], self.groups[i]))]

        self._unique_structures = [i[0] for i in self.groups]
        for s in structures:
            if s in self._unique_structures:
                self.structure_is_unique.append(True)
            else:
                self.structure_is_unique.append(False)
        self._not_duplicate = self.structure_is_unique

        if self.against_icsd:
            structure_file = "oqmd1.2_exp_based_entries_structures.json"
            cache_matrio_data(structure_file)
            with open(os.path.join(CAMD_CACHE, structure_file), 'r') as f:
                icsd_structures = json.load(f)
            chemsys = set()
            for s in self._unique_structures:
                chemsys = chemsys.union( set(s.composition.as_dict().keys()))

            self.icsd_structs_inchemsys = []
            for k, v in icsd_structures.items():
                try:
                    s = Structure.from_dict(v)
                    elems = set(s.composition.as_dict().keys())
                    if elems == chemsys:
                        self.icsd_structs_inchemsys.append(s)
                except:
                    warnings.warn("Unable to process structure {}".format(k))

            self.matching_icsd_strs = []
            for i in range(len(structures)):
                if self.structure_is_unique[i]:
                    match = None
                    for s2 in self.icsd_structs_inchemsys:
                        if smatch.fit(self.structures[i], s2):
                            match = s2
                            break
                    self.matching_icsd_strs.append(match) # store the matching ICSD structures.
                else:
                    self.matching_icsd_strs.append(None)

            # Flip matching bools, and create a filter
            self._icsd_filter = [not i for i in self.matching_icsd_strs]
            self.structure_is_unique = (np.array(self.structure_is_unique)
                                        * np.array(self._icsd_filter)).tolist()
            self.unique_structures = list(itertools.compress(
                self.structures, self.structure_is_unique))
        else:
            self.unique_structures = self._unique_structures

        # We store the final list of unique structures as unique_structures.
        # We return a corresponding list of bool to the initial structure
        # list provided.
        return self.structure_is_unique

    def analyze_vaspqmpy_jobs(self, jobs, against_icsd=False,
                              use_energies=False):
        """
        Useful for analysis integrated as part of a campaign itself

        Args:
            jobs:
            against_icsd:

        Returns:
        """
        self.structure_ids = []
        self.structures = []
        self.energies = []
        for j, r in jobs.items():
            if r['status'] == 'SUCCEEDED':
                self.structures.append(r['result']['output']['crystal'])
                self.structure_ids.append(j)
                self.energies.append(r['result']['output']['final_energy_per_atom'])
        if use_energies:
            return self.analyze(self.structures, self.structure_ids,
                                against_icsd, self.energies)
        else:
            return self.analyze(self.structures, self.structure_ids, against_icsd)

    def present(self):
        pass


class FinalizeQqmdCampaign:
    def __init__(self, hull_distance=None):
        self.hull_distance = hull_distance if hull_distance else 0.2
        self.path = None

    def finalize(self, path=None):
        """
        post-processing for dft-campaigns
        """
        self.path = path if path else '.'
        update_run_w_structure(self.path, hull_distance=self.hull_distance)


class AnalyzeStability(AnalyzerBase):
    def __init__(self, df=None, new_result_ids=None, hull_distance=None,
                 multiprocessing=True, entire_space=False):
        self.df = df
        self.new_result_ids = new_result_ids
        self.hull_distance = hull_distance if hull_distance else 0.05
        self.multiprocessing = multiprocessing
        self.entire_space = entire_space
        self.space = None
        self.stabilities = None
        super(AnalyzeStability, self).__init__()

    def filter_dataframe_by_composition(self, elements, df=None):
        """
        Filters dataframe by composition
        """
        df = df if df is not None else self.df
        elements = set(elements)
        ind_to_include = []
        for ind in self.df.index:
            if set(Composition(df.loc[ind]['Composition']).as_dict().keys()).issubset(elements):
                ind_to_include.append(ind)
        return df.loc[ind_to_include]

    def get_phase_space(self, df=None):
        """
        Gets PhaseSpace object associated with dataframe
        """
        _df = df if df is not None else self.df
        phases = []
        for data in _df.iterrows():
            phases.append(Phase(data[1]['Composition'], energy=data[1]['delta_e'],
                                per_atom=True, description=data[0]))
        for el in ELEMENTS:
            phases.append(Phase(el, 0.0, per_atom=True))

        pd = PhaseData()
        pd.add_phases(phases)
        space = PhaseSpaceAL(bounds=ELEMENTS, data=pd)
        return space

    def analyze(self, df=None, new_result_ids=None, all_result_ids=None,
                return_within_hull=True):
        """
        Args:
            df (DataFrame): data frame with structure-data for formation
                energy, composition, etc.
            new_result_ids (list): list of ids from the dataframe index
                corresponding to new results
            all_result_ids (list): list of ids from the dataframe index
                corresponding to all desired analyzed results
            return_within_hull (bool): whether to return boolean array
                corresponding to whether stabilities are within hull
                or raw results

        Returns:

        """
        include_columns = ['Composition', 'delta_e']
        self.df = df[include_columns].drop_duplicates(keep='last').dropna()
        # Note some of id's in all_result_ids may not have corresponding
        # experiment, if those exps. failed.
        self.all_result_ids = all_result_ids
        self.new_result_ids = new_result_ids

        if not self.entire_space:
            # Constrains the phase space to that of the target compounds.
            # More efficient when searching in a specified chemistry,
            # less efficient if larger spaces are without specified chemistry.
            comps = self.df.loc[all_result_ids]['Composition'].dropna()
            system_elements = []
            for comp in comps:
                system_elements += list(Composition(comp).as_dict().keys())
            # TODO: Fix this line to be compatible with
            # later versions of pandas (i.e. b/c all_result_ids may contain
            # things not in df currently (b/c of failed experiments).
            # We should test comps = self.df.loc[self.df.index.intersection(all_result_ids)]
            _df = self.filter_dataframe_by_composition(system_elements)
        else:
            _df = self.df

        space = self.get_phase_space(_df)

        if all_result_ids is not None:
            all_new_phases = [p for p in space.phases if p.description in all_result_ids]
        else:
            all_new_phases = None

        if self.multiprocessing:
            space.compute_stabilities_multi(phases_to_evaluate=all_new_phases)
        else:
            space.compute_stabilities_mod(phases_to_evaluate=all_new_phases)

        self.space = space

        # Key stabilities by ID
        stabilities_by_id = {phase.description: phase.stability
                             for phase in all_new_phases}
        self.stabilities = stabilities_by_id

        # Get stabilities of new and all ids
        stabilities_of_new = np.array([stabilities_by_id.get(uid, np.nan)
                                       for uid in self.new_result_ids], dtype=np.float)
        stabilities_of_all = np.array([stabilities_by_id.get(uid, np.nan)
                                       for uid in self.all_result_ids], dtype=np.float)

        # Cast to boolean if specified
        if return_within_hull:
            stabilities_of_new = stabilities_of_new <= self.hull_distance
            stabilities_of_all = stabilities_of_all <= self.hull_distance

        return stabilities_of_new, stabilities_of_all

    def present(self, df=None, new_result_ids=None, all_result_ids=None,
                filename=None, save_hull_distance=False, finalize=False):
        """
        Generate plots of convex hulls for each of the runs

        Args:
            df (DataFrame): dataframe with formation energies, compositions, ids
            new_result_ids ([]): list of new result ids (i. e. indexes
                in the updated dataframe)
            all_result_ids ([]): list of all result ids associated
                with the current run
            filename (str): filename to output, if None, no file output
                is produced

        Returns:
            (pyplot): plotter instance
        """
        df = df if df is not None else self.df
        new_result_ids = new_result_ids if new_result_ids is not None \
            else self.new_result_ids
        all_result_ids = all_result_ids if all_result_ids is not None \
            else self.all_result_ids

        # TODO: consolidate duplicated code here
        # Generate all entries
        comps = df.loc[all_result_ids]['Composition'].dropna()
        system_elements = []
        for comp in comps:
            system_elements += list(Composition(comp).as_dict().keys())
        elems = set(system_elements)
        if len(elems) > 4:
            warnings.warn("Number of elements too high for phase diagram plotting")
            return None
        ind_to_include = []
        for ind in df.index:
            if set(Composition(df.loc[ind]['Composition']).as_dict().keys()).issubset(elems):
                ind_to_include.append(ind)
        _df = df.loc[ind_to_include]

        # Create computed entry column
        _df['entry'] = [
            ComputedEntry(
                Composition(row['Composition']),
                row['delta_e'] * Composition(row['Composition']).num_atoms,  # un-normalize the energy
                entry_id=index
            )
            for index, row in _df.iterrows()]
        # Partition ids into sets of prior to CAMD run, from CAMD but prior to
        # current iteration, and new ids
        ids_prior_to_camd = list(set(_df.index) - set(all_result_ids))
        ids_prior_to_run = list(set(all_result_ids) - set(new_result_ids))

        # Create phase diagram based on everything prior to current run
        entries = list(_df.loc[ids_prior_to_run + ids_prior_to_camd]['entry'])
        # Filter for nans by checking if it's a computed entry
        entries = [entry for entry in entries if isinstance(entry, ComputedEntry)]
        pg_elements = [Element(el) for el in sorted(elems)]
        pd = PhaseDiagram(entries, elements=pg_elements)
        plotkwargs = {
            "markerfacecolor": "white",
            "markersize": 7,
            "linewidth": 2,
        }
        if finalize:
            plotkwargs.update({'linestyle': '--'})
        else:
            plotkwargs.update({'linestyle': '-'})
        plotter = PDPlotter(pd, **plotkwargs)

        getplotkwargs = {"label_stable": False} if finalize else {}
        plot = plotter.get_plot(**getplotkwargs)
        # Get valid results
        valid_results = [new_result_id for new_result_id in new_result_ids
                         if new_result_id in _df.index]

        if finalize:
            # If finalize, we'll reset pd to all entries at this point
            # to measure stabilities wrt. the ultimate hull.
            pd = PhaseDiagram(_df['entry'].values, elements=pg_elements)
            plotter = PDPlotter(pd, **{"markersize": 0, "linestyle": "-", "linewidth": 2})
            plot = plotter.get_plot(plt=plot)

        for entry in _df['entry'][valid_results]:
            decomp, e_hull = pd.get_decomp_and_e_above_hull(
                    entry, allow_negative=True)
            if e_hull < self.hull_distance:
                color = 'g'
                marker = 'o'
                markeredgewidth = 1
            else:
                color = 'r'
                marker = 'x'
                markeredgewidth = 1

            # Get coords
            coords = [entry.composition.get_atomic_fraction(el)
                      for el in pd.elements][1:]
            if pd.dim == 2:
                coords = coords + [pd.get_form_energy_per_atom(entry)]
            if pd.dim == 3:
                coords = triangular_coord(coords)
            elif pd.dim == 4:
                coords = tet_coord(coords)
            plot.plot(*coords, marker=marker, markeredgecolor=color,
                      markerfacecolor="None", markersize=11,
                      markeredgewidth=markeredgewidth)

        if filename is not None:
            plot.savefig(filename, dpi=70)
        plot.close()

        if filename is not None and save_hull_distance:
            if self.stabilities is None:
                print("ERROR: No stability information in analyzer.")
                return None
            with open(filename.split(".")[0]+'.json', 'w') as f:
                json.dump(self.stabilities, f)


class PhaseSpaceAL(PhaseSpace):
    """
    Modified qmpy.PhaseSpace for GCLP based stabiltiy computations
    TODO: basic multithread or Gurobi for gclp
    """

    def compute_stabilities_mod(self, phases_to_evaluate=None):
        """
        Calculate the stability for every Phase.

        Args:
            phases_to_evaluate ([phase]): Included phases, if None,
                uses every Phase in PhaseSpace.phases
        """

        if phases_to_evaluate is None:
            phases_to_evaluate = self.phases

        for p in tqdm(list(self.phase_dict.values())):
            if p.stability is None:  # for low e phases, we only need to eval stability if it doesn't exist
                try:
                    p.stability = p.energy - self.gclp(p.unit_comp)[0]
                except:
                    print(p)
                    p.stability = np.nan

        # will only do requested phases for things not in phase_dict
        for p in tqdm(phases_to_evaluate):
            if p not in list(self.phase_dict.values()):
                if p.name in self.phase_dict:
                    p.stability = p.energy - self.phase_dict[p.name].energy + self.phase_dict[p.name].stability
                else:
                    try:
                        p.stability = p.energy - self.gclp(p.unit_comp)[0]
                    except:
                        print(p)
                        p.stability = np.nan

    def compute_stabilities_multi(self, phases_to_evaluate=None,
                                  ncpus=multiprocessing.cpu_count()):
        """
        Calculate the stability for every Phase using multiprocessing

        Args:
            phases_to_evaluate ([phase]): Included phases, if None,
                uses every Phase in PhaseSpace.phases
            ncpus (int): number of cpus to use in multiprocessing
        """

        if phases_to_evaluate is None:
            phases_to_evaluate = self.phases

        # Creating a map from entry uid to index of entry in the
        # current list of phases in space.
        self.uid_to_phase_ind = dict([(self.phases[i].description, i)
                                      for i in range(len(self.phases))])

        phase_dict_list = list(self.phase_dict.values())
        _result_list1 = parmap(self._multiproc_help1,
                               phase_dict_list, nprocs=ncpus)
        for i in range(len(phase_dict_list)):
            self.phase_dict[phase_dict_list[i].name].stability = _result_list1[i]

        _result_list2 = parmap(self._multiproc_help2,
                               phases_to_evaluate, nprocs=ncpus)
        for i in range(len(phases_to_evaluate)):
            # we will use the uid_to_phase_ind create above to be able
            # to map results of parmap to self.phases
            ind = self.uid_to_phase_ind[phases_to_evaluate[i].description]
            self.phases[ind].stability = _result_list2[i]


    def _multiproc_help1(self, p):
        # For low e phases, we only eval stability if it doesn't exist
        if p.stability is None:
            try:
                p.stability = p.energy - self.gclp(p.unit_comp)[0]
            except:
                print(p)
                p.stability = np.nan
        return p.stability

    def _multiproc_help2(self, p):
        if p not in list(self.phase_dict.values()):
            if p.name in self.phase_dict:
                p.stability = p.energy - self.phase_dict[p.name].energy + self.phase_dict[p.name].stability
            else:
                try:
                    p.stability = p.energy - self.gclp(p.unit_comp)[0]
                except:
                    print(p)
                    p.stability = np.nan
        return p.stability


def fun(f, q_in, q_out):
    while True:
        i, x = q_in.get()
        if i is None:
            break
        q_out.put((i, f(x)))


def parmap(f, X, nprocs=multiprocessing.cpu_count()):
    q_in = multiprocessing.Queue(1)
    q_out = multiprocessing.Queue()

    proc = [multiprocessing.Process(target=fun, args=(f, q_in, q_out))
            for _ in range(nprocs)]
    for p in proc:
        p.daemon = True
        p.start()

    sent = [q_in.put((i, x)) for i, x in enumerate(X)]
    [q_in.put((None, None)) for _ in range(nprocs)]
    res = [q_out.get() for _ in range(len(sent))]

    [p.join() for p in proc]

    return [x for i, x in sorted(res)]


def update_run_w_structure(folder, hull_distance=0.2):
    """
    Updates a campaign grouped in directories with structure analysis

    """
    with cd(folder):
        required_files = ["seed_data.pickle"]
        if os.path.isfile("error.json"):
            error = loadfn("error.json")
            print("{} ERROR: {}".format(folder, error))

        if not all([os.path.isfile(fn) for fn in required_files]):
            print("{} ERROR: no seed data, no analysis to be done")
        else:
            iteration = -1
            jobs = {}
            while True:
                if os.path.isdir(str(iteration)):
                    jobs.update(loadfn(os.path.join(
                        str(iteration), '_exp_raw_results.json')))
                    iteration += 1
                else:
                    break
            with open("seed_data.pickle", "rb") as f:
                df = pickle.load(f)

            all_ids = loadfn("consumed_candidates.json")
            st_a = AnalyzeStability(df=df, hull_distance=hull_distance)
            _, stablities_of_discovered = st_a.analyze(df, all_ids, all_ids)

            # Having calculated stabilities again, we plot the overall hull.
            st_a.present(df, all_ids, all_ids, filename="hull_finalized.png",
                         finalize=True, save_hull_distance=True)

            stable_discovered = list(itertools.compress(
                all_ids, stablities_of_discovered))
            s_a = AnalyzeStructures()
            s_a.analyze_vaspqmpy_jobs(jobs, against_icsd=True, use_energies=True)
            unique_s_dict = {}
            for i in range(len(s_a.structures)):
                if s_a.structure_is_unique[i] and \
                        (s_a.structure_ids[i] in stable_discovered):
                    unique_s_dict[s_a.structure_ids[i]] = s_a.structures[i]

            with open("discovered_unique_structures.json", "w") as f:
                json.dump(dict([(k, s.as_dict())
                                for k, s in unique_s_dict.items()]), f)

            with open('structure_report.log', "w") as f:
                f.write("consumed discovery unique_discovery duplicate in_icsd \n")
                f.write(str(len(all_ids)) + ' ' +
                        str(len(stable_discovered)) + ' ' +
                        str(len(unique_s_dict)) + ' '
                        + str(len(s_a.structures) - sum(s_a._not_duplicate)) + ' '
                        + str(sum([not i for i in s_a._icsd_filter])))
